# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

# cython: language_level = 3


from libc.stdint cimport uintptr_t, int64_t
from cpython.pycapsule cimport PyCapsule_GetPointer

from nanoarrow_c cimport (
    ArrowArray,
    ArrowArrayStream,
    ArrowArrayStreamGetNext,
    ArrowArrayStreamGetSchema,
    ArrowArrayStreamMove,
    ArrowBasicArrayStreamInit,
    ArrowBasicArrayStreamSetArray,
    ArrowBasicArrayStreamValidate,
    ArrowBufferAppendInt64,
    ArrowResolveChunk64,
    ArrowType,
)

from nanoarrow cimport _types
from nanoarrow._array cimport CArray
from nanoarrow._buffer cimport CBuffer
from nanoarrow._schema cimport CSchema, assert_type_equal
from nanoarrow._utils cimport (
    alloc_c_array_stream,
    c_array_shallow_copy,
    Error
)

from typing import Iterable, List, Tuple

from nanoarrow import _repr_utils


cdef class CArrayStream:
    """Low-level ArrowArrayStream wrapper

    This object is a literal wrapper around an ArrowArrayStream. It provides methods that
    that wrap the underlying C callbacks and handles the C Data interface lifecycle
    (i.e., initialized ArrowArrayStream structures are always released).

    See `nanoarrow.c_array_stream()` for construction and usage examples.
    """
    cdef object _base
    cdef ArrowArrayStream* _ptr
    cdef object _cached_schema

    def __cinit__(self, object base, uintptr_t addr):
        self._base = base
        self._ptr = <ArrowArrayStream*>addr
        self._cached_schema = None

    @staticmethod
    def allocate() -> CArrayStream:
        """Allocate a released ArrowArrayStream"""
        cdef ArrowArrayStream* c_array_stream_out
        base = alloc_c_array_stream(&c_array_stream_out)
        return CArrayStream(base, <uintptr_t>c_array_stream_out)

    @staticmethod
    def from_c_arrays(arrays: List[CArray], CSchema schema, move=False, validate=True) -> CArrayStream:
        """Create an ArrowArrayStream from an existing set of arrays

        Given a previously resolved list of arrays, create an ArrowArrayStream
        representation of the sequence of chunks.

        Parameters
        ----------
        arrays : List[CArray]
            A list of arrays to use as batches.
        schema : CSchema
            The schema that will be returned. Must be type equal with the schema
            of each array (this is checked if validate is ``True``)
        move : bool, optional
            If True, transfer ownership from each array instead of creating a
            shallow copy. This is only safe if the caller knows the origin of the
            arrays and knows that they will not be accessed after this stream has been
            created.
        validate : bool, optional
            If True, enforce type equality between the provided schema and the schema
            of each array.
        """
        cdef ArrowArrayStream* c_array_stream_out
        base = alloc_c_array_stream(&c_array_stream_out)

        # Don't create more copies than we have to (but make sure
        # one exists for validation if requested)
        cdef CSchema out_schema = schema
        if validate and not move:
            validate_schema = schema
            out_schema = schema.__deepcopy__()
        elif validate:
            validate_schema = schema.__deepcopy__()
            out_schema = schema
        elif not move:
            out_schema = schema.__deepcopy__()

        cdef int code = ArrowBasicArrayStreamInit(c_array_stream_out, out_schema._ptr, len(arrays))
        Error.raise_error_not_ok("ArrowBasicArrayStreamInit()", code)

        cdef ArrowArray tmp
        cdef CArray array
        for i in range(len(arrays)):
            array = arrays[i]

            if validate:
                assert_type_equal(array.schema, validate_schema, False)

            if not move:
                c_array_shallow_copy(array._base, array._ptr, &tmp)
                ArrowBasicArrayStreamSetArray(c_array_stream_out, i, &tmp)
            else:
                ArrowBasicArrayStreamSetArray(c_array_stream_out, i, array._ptr)

        cdef Error error = Error()
        if validate:
            code = ArrowBasicArrayStreamValidate(c_array_stream_out, &error.c_error)
            error.raise_message_not_ok("ArrowBasicArrayStreamValidate()", code)

        return CArrayStream(base, <uintptr_t>c_array_stream_out)

    def release(self):
        """Explicitly call the release callback of this stream"""
        if self.is_valid():
            self._ptr.release(self._ptr)

    @staticmethod
    def _import_from_c_capsule(stream_capsule) -> CArrayStream:
        """Import from a ArrowArrayStream PyCapsule.

        Parameters
        ----------
        stream_capsule : PyCapsule
            A valid PyCapsule with name 'arrow_array_stream' containing an
            ArrowArrayStream pointer.
        """
        return CArrayStream(
            stream_capsule,
            <uintptr_t>PyCapsule_GetPointer(stream_capsule, 'arrow_array_stream')
        )

    def __arrow_c_stream__(self, requested_schema=None):
        """
        Export the stream as an Arrow C stream PyCapsule.

        Parameters
        ----------
        requested_schema : PyCapsule | None
            A PyCapsule containing a C ArrowSchema representation of a requested
            schema. Not supported.

        Returns
        -------
        PyCapsule
        """
        self._assert_valid()

        if requested_schema is not None:
            raise NotImplementedError("requested_schema")

        cdef:
            ArrowArrayStream* c_array_stream_out

        array_stream_capsule = alloc_c_array_stream(&c_array_stream_out)
        ArrowArrayStreamMove(self._ptr, c_array_stream_out)
        return array_stream_capsule

    def _addr(self) -> int:
        """test to see if this causes a ci fail"""
        return <uintptr_t>self._ptr

    def is_valid(self) -> bool:
        """Check for a non-null and non-released underlying ArrowArrayStream"""
        return self._ptr != NULL and self._ptr.release != NULL

    def _assert_valid(self):
        if self._ptr == NULL:
            raise RuntimeError("array stream pointer is NULL")
        if self._ptr.release == NULL:
            raise RuntimeError("array stream is released")

    def _get_schema(self, CSchema schema):
        self._assert_valid()
        cdef Error error = Error()
        cdef int code = ArrowArrayStreamGetSchema(self._ptr, schema._ptr, &error.c_error)
        error.raise_message_not_ok("ArrowArrayStream::get_schema()", code)

    def _get_cached_schema(self):
        if self._cached_schema is None:
            self._cached_schema = CSchema.allocate()
            self._get_schema(self._cached_schema)

        return self._cached_schema

    def get_schema(self) -> CSchema:
        """Get the schema associated with this stream

        Calling this method will always issue a call to the underlying stream's
        get_schema callback.
        """
        out = CSchema.allocate()
        self._get_schema(out)
        return out

    def get_next(self) -> CArray:
        """Get the next Array from this stream

        Raises StopIteration when there are no more arrays in this stream.
        """
        self._assert_valid()

        # We return a reference to the same Python object for each
        # Array that is returned. This is independent of get_schema(),
        # which is guaranteed to call the C object's callback and
        # faithfully pass on the returned value.

        cdef Error error = Error()
        cdef CArray array = CArray.allocate(self._get_cached_schema())
        cdef int code = ArrowArrayStreamGetNext(self._ptr, array._ptr, &error.c_error)
        error.raise_message_not_ok("ArrowArrayStream::get_next()", code)

        if not array.is_valid():
            raise StopIteration()
        else:
            return array

    def __iter__(self):
        return self

    def __next__(self):
        return self.get_next()

    def __enter__(self):
        return self

    def __exit__(self, *args, **kwargs):
        self.release()

    def __repr__(self):
        return _repr_utils.array_stream_repr(self)


cdef class CMaterializedArrayStream:
    """Optimized representation of a fully consumed ArrowArrayStream

    This class provides a data structure similar to pyarrow's ChunkedArray
    where each consumed array is a referenced-counted shared array. This
    class wraps the utilities provided by the nanoarrow C library to iterate
    over and facilitate log(n) random access to items in this container.
    """
    cdef CSchema _schema
    cdef CBuffer _array_ends
    cdef list _arrays
    cdef int64_t _total_length

    def __cinit__(self):
        self._arrays = []
        self._total_length = 0
        self._schema = CSchema.allocate()
        self._array_ends = CBuffer.empty()
        cdef int code = ArrowBufferAppendInt64(self._array_ends._ptr, 0)
        Error.raise_error_not_ok("ArrowBufferAppendInt64()", code)

    cdef _finalize(self):
        self._array_ends._set_data_type(<ArrowType>_types.INT64)

    @property
    def schema(self) -> CSchema:
        return self._schema

    def __getitem__(self, k) -> Tuple[CArray, int]:
        cdef int64_t kint
        cdef int array_i
        cdef const int64_t* sorted_offsets = <int64_t*>self._array_ends._ptr.data

        if isinstance(k, slice):
            raise NotImplementedError("index with slice")

        kint = k
        if kint < 0:
            kint += self._total_length
        if kint < 0 or kint >= self._total_length:
            raise IndexError(f"Index {kint} is out of range")

        array_i = ArrowResolveChunk64(kint, sorted_offsets, 0, len(self._arrays))
        kint -= sorted_offsets[array_i]
        return self._arrays[array_i], kint

    def __len__(self) -> int:
        return self._array_ends[len(self._arrays)]

    def __iter__(self) -> Iterable[Tuple[CArray, int]]:
        for c_array in self._arrays:
            for item_i in range(len(c_array)):
                yield c_array, item_i

    def array(self, int64_t i) -> CArray:
        return self._arrays[i]

    @property
    def n_arrays(self) -> int:
        return len(self._arrays)

    @property
    def arrays(self) -> Iterable[CArray]:
        return iter(self._arrays)

    def __arrow_c_stream__(self, requested_schema=None):
        # When an array stream from iterable is supported, that could be used here
        # to avoid unnessary shallow copies.
        stream = CArrayStream.from_c_arrays(
            self._arrays,
            self._schema,
            move=False,
            validate=False
        )

        return stream.__arrow_c_stream__(requested_schema=requested_schema)

    def child(self, int64_t i) -> CMaterializedArrayStream:
        cdef CMaterializedArrayStream out = CMaterializedArrayStream()
        cdef int code

        out._schema = self._schema.child(i)
        out._arrays = [chunk.child(i) for chunk in self._arrays]
        for child_chunk in out._arrays:
            out._total_length += len(child_chunk)
            code = ArrowBufferAppendInt64(out._array_ends._ptr, out._total_length)
            Error.raise_error_not_ok("ArrowBufferAppendInt64()", code)

        out._finalize()
        return out

    @staticmethod
    def from_c_arrays(arrays: Iterable[CArray], CSchema schema, bint validate=True):
        """"Create a materialized array stream from an existing iterable of arrays

        This is slightly more efficient than creating a stream and then consuming it
        because the implementation can avoid a shallow copy of each array.
        """
        cdef CMaterializedArrayStream out = CMaterializedArrayStream()

        for array in arrays:
            if not isinstance(array, CArray):
                raise TypeError(f"Expected CArray but got {type(array).__name__}")

            if len(array) == 0:
                continue

            if validate:
                assert_type_equal(array.schema, schema, False)

            out._total_length += len(array)
            code = ArrowBufferAppendInt64(out._array_ends._ptr, out._total_length)
            Error.raise_error_not_ok("ArrowBufferAppendInt64()", code)
            out._arrays.append(array)

        out._schema = schema
        out._finalize()
        return out

    @staticmethod
    def from_c_array(CArray array) -> CMaterializedArrayStream:
        """"Create a materialized array stream from a single array
        """
        return CMaterializedArrayStream.from_c_arrays(
            [array],
            array.schema,
            validate=False
        )

    @staticmethod
    def from_c_array_stream(CArrayStream stream) -> CMaterializedArrayStream:
        """"Create a materialized array stream from an unmaterialized ArrowArrayStream
        """
        with stream:
            return CMaterializedArrayStream.from_c_arrays(
                stream,
                stream._get_cached_schema(),
                validate=False
            )



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from typing import Any, Callable, List, Sequence, Tuple, Union

from nanoarrow._array import CArrayView
from nanoarrow._buffer import CBuffer, CBufferBuilder
from nanoarrow.c_array_stream import c_array_stream
from nanoarrow.c_schema import c_schema_view
from nanoarrow.extension import resolve_extension
from nanoarrow.iterator import ArrayViewBaseIterator, PyIterator
from nanoarrow.schema import Type

from nanoarrow import _types


class ArrayViewVisitable:
    """Mixin class providing conversion methods based on visitors

    Can be used with classes that implement ``__arrow_c_stream__()``
    or ``__arrow_c_array__()``.
    """

    def to_pylist(self) -> List:
        """Convert to a ``list`` of Python objects

        Computes an identical value to ``list(iter_py())`` but can be much
        faster.

        Examples
        --------
        >>> import nanoarrow as na
        >>> from nanoarrow import visitor
        >>> array = na.Array([1, 2, 3], na.int32())
        >>> array.to_pylist()
        [1, 2, 3]
        """
        return ToPyListConverter.visit(self)

    def to_columns_pysequence(
        self, *, handle_nulls=None
    ) -> Tuple[List[str], List[Sequence]]:
        """Convert to a ``list`` of contiguous sequences

        Experimentally converts a stream of struct arrays into a list of contiguous
        sequences using the same logic as :meth:`to_pysequence`.

        Paramters
        ---------
        handle_nulls : callable
            A function returning a sequence based on a validity bytemap and a
            contiguous buffer of values. If the array contains no nulls, the
            validity bytemap will be ``None``. Built-in handlers include
            :func:`nulls_as_sentinel`, :func:`nulls_forbid`, and
            :func:`nulls_separate`). The default value is :func:`nulls_forbid`.

        Examples
        --------
        >>> import nanoarrow as na
        >>> import pyarrow as pa
        >>> batch = pa.record_batch({"col1": [1, 2, 3], "col2": ["a", "b", "c"]})
        >>> names, columns = na.Array(batch).to_columns_pysequence()
        >>> names
        ['col1', 'col2']
        >>> columns
        [nanoarrow.c_buffer.CBuffer(int64[24 b] 1 2 3), ['a', 'b', 'c']]
        """
        return ToColumnsPysequenceConverter.visit(self, handle_nulls=handle_nulls)

    def to_pysequence(self, *, handle_nulls=None) -> Sequence:
        """Convert to a contiguous sequence

        Experimentally converts a stream of arrays into a columnar representation
        such that each column is either a contiguous buffer or a ``list``.
        Integer, float, and interval arrays are currently converted to their
        contiguous buffer representation; other types are returned as a list
        of Python objects. The sequences returned by :meth:`to_pysequence` are
        designed to work as input to ``pandas.Series`` and/or ``numpy.array()``.
        The default conversions are subject to change based on initial user
        feedback.

        Parameters
        ----------
        handle_nulls : callable
            A function returning a sequence based on a validity bytemap and a
            contiguous buffer of values. If the array contains no nulls, the
            validity bytemap will be ``None``. Built-in handlers include
            :func:`nulls_as_sentinel`, :func:`nulls_forbid`, and
            :func:`nulls_separate`). The default value is :func:`nulls_forbid`.

        Examples
        --------
        >>> import nanoarrow as na
        >>> na.Array([1, 2, 3], na.int32()).to_pysequence()
        nanoarrow.c_buffer.CBuffer(int32[12 b] 1 2 3)
        """
        return ToPySequenceConverter.visit(self, handle_nulls=handle_nulls)


def nulls_forbid() -> Callable[[CBuffer, Sequence], Sequence]:
    """Erroring null handler

    A null handler that errors when it encounters nulls.

    Examples
    --------

    >>> import nanoarrow as na
    >>> na.Array([1, 2, 3], na.int32()).to_pysequence(handle_nulls=na.nulls_forbid())
    nanoarrow.c_buffer.CBuffer(int32[12 b] 1 2 3)
    >>> na.Array([1, None, 3], na.int32()).to_pysequence(handle_nulls=na.nulls_forbid())
    Traceback (most recent call last):
    ...
    ValueError: Null present with null_handler=nulls_forbid()
    """

    def handle(is_valid, data):
        # the is_valid bytemap is only created if there was at least one null
        if is_valid is not None:
            raise ValueError("Null present with null_handler=nulls_forbid()")

        return data

    return handle


def nulls_as_sentinel(sentinel=None):
    """Sentinel null handler

    A null handler that assigns a sentinel to null values. This is
    done using numpy using the expression ``data[~is_valid] = sentinel``.
    The default sentinel value of ``None`` will result in float output and ``nan``
    assigned to null values for numeric and boolean inputs. This
    corresponds to numpy's handling of ``None`` in ``np.result_type()``
    and ``result[~is_valid] = None``.

    Parameters
    ----------
    sentinel : scalar, optional
        The value with which nulls should be replaced.

    Examples
    --------

    >>> import nanoarrow as na
    >>> na_array = na.Array([1, 2, 3], na.int32())
    >>> na_array.to_pysequence(handle_nulls=na.nulls_as_sentinel())
    array([1, 2, 3], dtype=int32)
    >>> na_array = na.Array([1, None, 3], na.int32())
    >>> na_array.to_pysequence(handle_nulls=na.nulls_as_sentinel())
    array([ 1., nan,  3.])
    >>> na_array.to_pysequence(handle_nulls=na.nulls_as_sentinel(-999))
    array([   1, -999,    3], dtype=int32)
    """
    import numpy as np

    def handle(is_valid, data):
        data = np.array(data, copy=False)

        if is_valid is not None:
            is_valid = np.array(is_valid, copy=False)
            out_type = np.result_type(data, sentinel)
            data = np.array(data, dtype=out_type, copy=True)
            data[~is_valid] = sentinel
            return data
        else:
            return data

    return handle


def nulls_separate() -> Callable[[CBuffer, Sequence], Tuple[CBuffer, Sequence]]:
    """Return nulls as a tuple of is_valid, data

    A null handler that returns its components.

    Examples
    --------

    >>> import nanoarrow as na
    >>> na_array = na.Array([1, 2, 3], na.int32())
    >>> na_array.to_pysequence(handle_nulls=na.nulls_separate())
    (None, nanoarrow.c_buffer.CBuffer(int32[12 b] 1 2 3))
    >>> na_array = na.Array([1, None, 3], na.int32())
    >>> result = na_array.to_pysequence(handle_nulls=na.nulls_separate())
    >>> result[0]
    nanoarrow.c_buffer.CBuffer(uint8[3 b] True False True)
    >>> result[1]
    nanoarrow.c_buffer.CBuffer(int32[12 b] 1 0 3)
    """

    def handle(is_valid, data):
        return is_valid, data

    return handle


class ArrayViewVisitor(ArrayViewBaseIterator):
    """Compute a value from one or more arrays as ArrowArrayViews

    This class supports a (currently internal) pattern for building
    output from a zero or more arrays in a stream.
    """

    @classmethod
    def visit(cls, obj, schema=None, total_elements=None, **kwargs):
        """Visit all chunks in ``obj`` as a :func:`c_array_stream`."""

        if total_elements is None and hasattr(obj, "__len__"):
            total_elements = len(obj)

        with c_array_stream(obj, schema=schema) as stream:
            visitor = cls(stream._get_cached_schema(), **kwargs)
            visitor.begin(total_elements)

            visitor_set_array = visitor._set_array
            visit_chunk_view = visitor.visit_chunk_view
            array_view = visitor._array_view

            for array in stream:
                visitor_set_array(array)
                visit_chunk_view(array_view)

        return visitor.finish()

    def begin(self, total_elements: Union[int, None] = None):
        """Called after the schema has been resolved but before any
        chunks have been visited. If the total number of elements
        (i.e., the sum of all chunk lengths) is known, it is provided here.
        """
        pass

    def visit_chunk_view(self, array_view: CArrayView) -> None:
        """Called exactly one for each chunk seen."""
        pass

    def finish(self) -> Any:
        """Called exactly once after all chunks have been visited."""
        return None


class ToPySequenceConverter(ArrayViewVisitor):
    def __init__(self, schema, handle_nulls=None, *, array_view=None):
        super().__init__(schema, array_view=array_view)
        cls, kwargs = _resolve_converter_cls(self._schema, handle_nulls=handle_nulls)
        self._visitor = cls(schema, **kwargs, array_view=self._array_view)

    def begin(self, total_elements: Union[int, None] = None):
        self._visitor.begin(total_elements)

    def visit_chunk_view(self, array_view: CArrayView) -> None:
        self._visitor.visit_chunk_view(array_view)

    def finish(self) -> Any:
        return self._visitor.finish()


class ToColumnsPysequenceConverter(ArrayViewVisitor):
    def __init__(self, schema, handle_nulls=None, *, array_view=None):
        super().__init__(schema, array_view=array_view)

        if self.schema.type != Type.STRUCT:
            raise ValueError("ToColumnListConverter can only be used on a struct array")

        # Resolve the appropriate visitor for each column
        self._child_visitors = []
        for child_schema, child_array_view in zip(
            self._schema.children, self._array_view.children
        ):
            self._child_visitors.append(
                self._resolve_child_visitor(
                    child_schema, child_array_view, handle_nulls
                )
            )

    def _resolve_child_visitor(self, child_schema, child_array_view, handle_nulls):
        cls, kwargs = _resolve_converter_cls(child_schema, handle_nulls)
        return cls(child_schema, **kwargs, array_view=child_array_view)

    def begin(self, total_elements: Union[int, None] = None) -> None:
        for child_visitor in self._child_visitors:
            child_visitor.begin(total_elements)

    def visit_chunk_view(self, array_view: CArrayView) -> Any:
        # This visitor does not handle nulls because it has no way to propagate these
        # into the child columns. It is designed to be used on top-level record batch
        # arrays which typically are marked as non-nullable or do not contain nulls.
        if array_view.null_count > 0:
            raise ValueError("null_count > 0 encountered in ToColumnListConverter")

        for child_visitor, child_array_view in zip(
            self._child_visitors, array_view.children
        ):
            child_visitor.visit_chunk_view(child_array_view)

    def finish(self) -> Tuple[List[str], List[Sequence]]:
        return [child.name for child in self._schema.children], [
            v.finish() for v in self._child_visitors
        ]


class ToPyListConverter(ArrayViewVisitor):
    def __init__(self, schema, *, iterator_cls=PyIterator, array_view=None):
        super().__init__(schema, array_view=array_view)

        # Ensure that self._iterator._array_view is self._array_view
        self._iterator = iterator_cls(schema, array_view=self._array_view)

    def begin(self, total_elements: Union[int, None] = None):
        self._lst = []

    def visit_chunk_view(self, array_view: CArrayView):
        # The constructor here ensured that self._iterator._array_view
        # is populated when self._set_array() is called.
        self._lst.extend(self._iterator)

    def finish(self) -> List:
        return self._lst


class ToPyBufferConverter(ArrayViewVisitor):
    def begin(self, total_elements: Union[int, None]):
        self._builder = self._make_builder()

        if total_elements is not None:
            element_size_bits = self._schema_view.layout.element_size_bits[1]
            element_size_bytes = element_size_bits // 8
            self._builder.reserve_bytes(total_elements * element_size_bytes)

    def visit_chunk_view(self, array_view: CArrayView) -> None:
        builder = self._builder
        offset, length = array_view.offset, array_view.length
        dst_bytes = length * builder.itemsize

        builder.reserve_bytes(dst_bytes)
        array_view.buffer(1).copy_into(builder, offset, length, len(builder))
        builder.advance(dst_bytes)

    def finish(self) -> Any:
        return self._builder.finish()

    def _make_builder(self):
        return CBufferBuilder().set_format(self._schema_view.buffer_format)


class ToBooleanBufferConverter(ArrayViewVisitor):
    def begin(self, total_elements: Union[int, None]):
        self._builder = CBufferBuilder()
        self._builder.set_format("?")
        if total_elements is not None:
            self._builder.reserve_bytes(total_elements)

    def visit_chunk_view(self, array_view: CArrayView) -> None:
        builder = self._builder
        offset, length = array_view.offset, array_view.length
        builder.reserve_bytes(length)
        array_view.buffer(1).unpack_bits_into(builder, offset, length, len(builder))
        builder.advance(length)

    def finish(self) -> Any:
        return self._builder.finish()


class ToNullableSequenceConverter(ArrayViewVisitor):
    def __init__(
        self,
        schema,
        converter_cls=ToPyBufferConverter,
        handle_nulls: Union[Callable[[CBuffer, Sequence], Any], None] = None,
        *,
        array_view=None
    ):
        super().__init__(schema, array_view=array_view)
        self._converter = converter_cls(schema, array_view=self._array_view)

        if handle_nulls is None:
            self._handle_nulls = nulls_forbid()
        else:
            self._handle_nulls = handle_nulls

    def begin(self, total_elements: Union[int, None]):
        self._builder = CBufferBuilder()
        self._builder.set_format("?")
        self._length = 0

        self._converter.begin(total_elements)

    def visit_chunk_view(self, array_view: CArrayView) -> None:
        offset, length = array_view.offset, array_view.length

        builder = self._builder
        chunk_contains_nulls = array_view.null_count != 0
        bitmap_allocated = len(builder) > 0

        if chunk_contains_nulls:
            current_length = self._length
            if not bitmap_allocated:
                builder.write_fill(1, current_length)

            builder.reserve_bytes(length)
            array_view.buffer(0).unpack_bits_into(
                builder, offset, length, current_length
            )
            builder.advance(length)

        elif bitmap_allocated:
            builder.write_fill(1, length)

        self._length += length
        self._converter.visit_chunk_view(array_view)

    def finish(self) -> Any:
        is_valid = self._builder.finish()
        data = self._converter.finish()
        return self._handle_nulls(is_valid if len(is_valid) > 0 else None, data)


def _resolve_converter_cls(schema, handle_nulls=None):
    schema_view = c_schema_view(schema)
    ext = resolve_extension(schema_view)
    ext_converter_cls = ext.get_sequence_converter(schema) if ext else None

    if schema_view.nullable:
        if ext_converter_cls:
            return ToNullableSequenceConverter, {
                "converter_cls": ext_converter_cls,
                "handle_nulls": handle_nulls,
            }
        elif schema_view.type_id == _types.BOOL:
            return ToNullableSequenceConverter, {
                "converter_cls": ToBooleanBufferConverter,
                "handle_nulls": handle_nulls,
            }
        elif schema_view.buffer_format is not None:
            return ToNullableSequenceConverter, {
                "converter_cls": ToPyBufferConverter,
                "handle_nulls": handle_nulls,
            }
        else:
            return ToPyListConverter, {}
    else:
        if ext_converter_cls:
            return ext_converter_cls, {}
        elif schema_view.type_id == _types.BOOL:
            return ToBooleanBufferConverter, {}
        elif schema_view.buffer_format is not None:
            return ToPyBufferConverter, {}
        else:
            return ToPyListConverter, {}



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

import pytest

import nanoarrow as na


def test_buffer_view_bool_():
    bool_array_view = na.c_array([1, 0, 0, 1], na.bool_()).view()
    view = bool_array_view.buffer(1)

    assert view.element_size_bits == 1
    assert view.size_bytes == 1
    assert view.data_type_id == na.Type.BOOL.value
    assert view.data_type == "bool"
    assert view.format == "B"

    # Check item interface
    assert len(view) == 1
    assert view[0] == 0b1001
    assert list(view) == [0b1001]

    # Check against buffer protocol
    mv = memoryview(view)
    assert len(mv) == len(view)
    assert mv[0] == view[0]
    assert list(mv) == list(view)

    # Check element interface
    assert view.n_elements == 8
    assert list(view.elements()) == [True, False, False, True] + [False] * 4
    assert [view.element(i) for i in range(8)] == list(view.elements())

    # Check element slices
    assert list(view.elements(0, 4)) == [True, False, False, True]
    assert list(view.elements(1, 3)) == [False, False, True]

    msg = "do not describe a valid slice"
    with pytest.raises(IndexError, match=msg):
        view.elements(-1, None)
    with pytest.raises(IndexError, match=msg):
        view.elements(0, -1)
    with pytest.raises(IndexError, match=msg):
        view.elements(0, 9)

    # Check repr
    assert "10010000" in repr(view)


def test_buffer_view_bool_unpack():
    from array import array

    bool_array_view = na.c_array([1, 0, 0, 1], na.bool_()).view()
    view = bool_array_view.buffer(1)

    # Check unpacking
    unpacked_all = view.unpack_bits()
    assert len(unpacked_all) == view.n_elements
    assert unpacked_all.data_type == "uint8"
    assert unpacked_all.format == "?"
    assert list(unpacked_all) == [1, 0, 0, 1, 0, 0, 0, 0]

    unpacked_some = view.unpack_bits(1, 4)
    assert len(unpacked_some) == 4
    assert list(unpacked_some) == [0, 0, 1, 0]

    # Check with non-zero destination offset
    out = bytearray([255] * 10)
    assert view.unpack_bits_into(out, dest_offset=2) == 8
    assert list(out) == [255, 255, 1, 0, 0, 1, 0, 0, 0, 0]

    # Check error requesting out-of-bounds dest_offset
    with pytest.raises(IndexError, match="Can't unpack"):
        view.unpack_bits_into(out, dest_offset=-1)

    # Check errors from requesting out-of-bounds slices
    msg = "do not describe a valid slice"
    with pytest.raises(IndexError, match=msg):
        view.unpack_bits(-1, None)
    with pytest.raises(IndexError, match=msg):
        view.unpack_bits(0, -1)
    with pytest.raises(IndexError, match=msg):
        view.unpack_bits(0, 9)

    # Check errors from an output buffer of insufficient length
    out = bytearray()
    msg = "Can't unpack 8 elements into buffer of size 0"
    with pytest.raises(IndexError, match=msg):
        view.unpack_bits_into(out)

    # Check errors from an output buffer with the wrong data type
    out = array("i", [0, 0, 0, 0])
    msg = "Destination buffer must have itemsize == 1"
    with pytest.raises(ValueError, match=msg):
        view.unpack_bits_into(out)


def test_buffer_view_non_bool():
    array_view = na.c_array([1, 2, 3, 5], na.int32()).view()
    view = array_view.buffer(1)

    assert view.element_size_bits == 32
    assert view.size_bytes == 4 * 4
    assert view.data_type_id == na.Type.INT32.value
    assert view.data_type == "int32"
    assert view.format == "i"

    # Check item interface
    assert len(view) == 4
    assert list(view) == [1, 2, 3, 5]
    assert [view[i] for i in range(4)] == list(view)

    # Check against buffer protocol
    mv = memoryview(view)
    assert len(mv) == len(view)
    assert mv[0] == view[0]
    assert [mv[i] for i in range(4)] == [view[i] for i in range(4)]

    # Check element interface
    assert view.n_elements == len(view)
    assert list(view.elements()) == list(view)
    assert [view.element(i) for i in range(4)] == list(view.elements())

    # Check element slices
    assert list(view.elements(0, 3)) == [1, 2, 3]
    assert list(view.elements(1, 3)) == [2, 3, 5]

    with pytest.raises(IndexError, match="do not describe a valid slice"):
        view.elements(-1, None)
    with pytest.raises(IndexError, match="do not describe a valid slice"):
        view.elements(0, -1)
    with pytest.raises(IndexError, match="do not describe a valid slice"):
        view.elements(1, 4)

    # Check that unpacking will error
    with pytest.raises(ValueError, match="Can't unpack non-boolean buffer"):
        view.unpack_bits()

    # Check repr
    assert "1 2 3 5" in repr(view)


def test_buffer_view_copy():
    from array import array

    array_view = na.c_array([1, 2, 3, 4], na.int32()).view()
    view = array_view.buffer(1)

    # Check copying
    copied_all = view.copy()
    assert len(copied_all) == view.n_elements
    assert copied_all.data_type == "int32"
    assert list(copied_all) == [1, 2, 3, 4]

    copied_some = view.copy(1, 3)
    assert len(copied_some) == 3
    assert list(copied_some) == [2, 3, 4]

    # Check with non-zero destination offset
    out = array(view.format, [0, 0, 0, 0, 0, 0])
    assert view.copy_into(out, dest_offset=2) == 16
    assert list(out) == [0, 0, 1, 2, 3, 4]

    # Check error requesting out-of-bounds dest_offset
    with pytest.raises(IndexError, match="Can't unpack"):
        view.copy_into(out, dest_offset=-1)

    # Check errors from requesting out-of-bounds slices
    msg = "do not describe a valid slice"
    with pytest.raises(IndexError, match=msg):
        view.copy(-1, None)
    with pytest.raises(IndexError, match=msg):
        view.copy(0, -1)
    with pytest.raises(IndexError, match=msg):
        view.copy(0, 9)

    # Check errors from an output buffer of insufficient length
    out = array("i")
    msg = "Can't unpack 4 elements into buffer of size 0"
    with pytest.raises(IndexError, match=msg):
        view.copy_into(out)

    # Check errors from an output buffer with the wrong data type
    out = array("d", [0, 0, 0, 0])
    msg = "Destination buffer must have itemsize == 1 or itemsize == 4"
    with pytest.raises(ValueError, match=msg):
        view.copy_into(out)



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

import pytest
from nanoarrow.c_schema import c_schema_view

import nanoarrow as na
from nanoarrow import extension


def test_basic_extension():
    class TestExtension(extension.Extension):
        def get_schema(self):
            return na.extension_type(na.int32(), "arrow.test")

        def get_params(self, c_schema):
            return {"parsed_key": "some parsed value"}

    instance = TestExtension()
    assert extension.register_extension(instance) is None

    # Check internal resolution
    assert extension.resolve_extension(c_schema_view(instance.get_schema())) is instance

    # Check Schema integration
    schema = na.extension_type(na.int32(), "arrow.test")
    assert schema.extension.parsed_key == "some parsed value"

    # Ensure other integrations fail if methods aren't implemented
    with pytest.raises(TypeError, match="get_iterable_appender"):
        assert na.Array([0], schema)

    with pytest.raises(TypeError, match="get_buffer_appender"):
        assert na.Array(bytearray([0]), schema)

    schema = na.extension_type(na.int32(), "arrow.test")
    storage_array = na.c_array([1, 2, 3], na.int32())
    _, storage_array_capsule = na.c_array(storage_array).__arrow_c_array__()
    array = na.Array(storage_array_capsule, schema)
    with pytest.raises(NotImplementedError, match="get_pyiter"):
        array.to_pylist()

    with pytest.raises(NotImplementedError, match="get_sequence_converter"):
        array.to_pysequence()

    other_instance = TestExtension()
    assert extension.register_extension(other_instance) is instance
    assert extension.unregister_extension("arrow.test") is other_instance
    with pytest.raises(KeyError):
        extension.unregister_extension("arrow.test")



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

import _cython_3_0_11
import enum
from typing import Callable, ClassVar

BINARY: CArrowType
BINARY_VIEW: CArrowType
BOOL: CArrowType
DATE32: CArrowType
DATE64: CArrowType
DECIMAL128: CArrowType
DECIMAL256: CArrowType
DENSE_UNION: CArrowType
DICTIONARY: CArrowType
DOUBLE: CArrowType
DURATION: CArrowType
EXTENSION: CArrowType
FIXED_SIZE_BINARY: CArrowType
FIXED_SIZE_LIST: CArrowType
FLOAT: CArrowType
HALF_FLOAT: CArrowType
INT16: CArrowType
INT32: CArrowType
INT64: CArrowType
INT8: CArrowType
INTERVAL_DAY_TIME: CArrowType
INTERVAL_MONTHS: CArrowType
INTERVAL_MONTH_DAY_NANO: CArrowType
LARGE_BINARY: CArrowType
LARGE_LIST: CArrowType
LARGE_STRING: CArrowType
LIST: CArrowType
MAP: CArrowType
NA: CArrowType
RUN_END_ENCODED: CArrowType
SPARSE_UNION: CArrowType
STRING: CArrowType
STRING_VIEW: CArrowType
STRUCT: CArrowType
TIME32: CArrowType
TIME64: CArrowType
TIMESTAMP: CArrowType
UINT16: CArrowType
UINT32: CArrowType
UINT64: CArrowType
UINT8: CArrowType
UNINITIALIZED: CArrowType
__pyx_capi__: dict
__test__: dict
has_time_unit: _cython_3_0_11.cython_function_or_method
is_data_view: _cython_3_0_11.cython_function_or_method
is_decimal: _cython_3_0_11.cython_function_or_method
is_fixed_size: _cython_3_0_11.cython_function_or_method
is_floating_point: _cython_3_0_11.cython_function_or_method
is_signed_integer: _cython_3_0_11.cython_function_or_method
is_union: _cython_3_0_11.cython_function_or_method
is_unsigned_integer: _cython_3_0_11.cython_function_or_method
sys_byteorder: str

class CArrowType(enum.IntFlag):
    __new__: ClassVar[Callable] = ...
    BINARY: ClassVar[CArrowType] = ...
    BINARY_VIEW: ClassVar[CArrowType] = ...
    BOOL: ClassVar[CArrowType] = ...
    DATE32: ClassVar[CArrowType] = ...
    DATE64: ClassVar[CArrowType] = ...
    DECIMAL128: ClassVar[CArrowType] = ...
    DECIMAL256: ClassVar[CArrowType] = ...
    DENSE_UNION: ClassVar[CArrowType] = ...
    DICTIONARY: ClassVar[CArrowType] = ...
    DOUBLE: ClassVar[CArrowType] = ...
    DURATION: ClassVar[CArrowType] = ...
    EXTENSION: ClassVar[CArrowType] = ...
    FIXED_SIZE_BINARY: ClassVar[CArrowType] = ...
    FIXED_SIZE_LIST: ClassVar[CArrowType] = ...
    FLOAT: ClassVar[CArrowType] = ...
    HALF_FLOAT: ClassVar[CArrowType] = ...
    INT16: ClassVar[CArrowType] = ...
    INT32: ClassVar[CArrowType] = ...
    INT64: ClassVar[CArrowType] = ...
    INT8: ClassVar[CArrowType] = ...
    INTERVAL_DAY_TIME: ClassVar[CArrowType] = ...
    INTERVAL_MONTHS: ClassVar[CArrowType] = ...
    INTERVAL_MONTH_DAY_NANO: ClassVar[CArrowType] = ...
    LARGE_BINARY: ClassVar[CArrowType] = ...
    LARGE_LIST: ClassVar[CArrowType] = ...
    LARGE_STRING: ClassVar[CArrowType] = ...
    LIST: ClassVar[CArrowType] = ...
    MAP: ClassVar[CArrowType] = ...
    NA: ClassVar[CArrowType] = ...
    RUN_END_ENCODED: ClassVar[CArrowType] = ...
    SPARSE_UNION: ClassVar[CArrowType] = ...
    STRING: ClassVar[CArrowType] = ...
    STRING_VIEW: ClassVar[CArrowType] = ...
    STRUCT: ClassVar[CArrowType] = ...
    TIME32: ClassVar[CArrowType] = ...
    TIME64: ClassVar[CArrowType] = ...
    TIMESTAMP: ClassVar[CArrowType] = ...
    UINT16: ClassVar[CArrowType] = ...
    UINT32: ClassVar[CArrowType] = ...
    UINT64: ClassVar[CArrowType] = ...
    UINT8: ClassVar[CArrowType] = ...
    UNINITIALIZED: ClassVar[CArrowType] = ...
    _all_bits_: ClassVar[int] = ...
    _boundary_: ClassVar[enum.FlagBoundary] = ...
    _flag_mask_: ClassVar[int] = ...
    _generate_next_value_: ClassVar[Callable] = ...
    _inverted_: ClassVar[None] = ...
    _member_map_: ClassVar[dict] = ...
    _member_names_: ClassVar[list] = ...
    _member_type_: ClassVar[type[int]] = ...
    _singles_mask_: ClassVar[int] = ...
    _unhashable_values_: ClassVar[list] = ...
    _use_args_: ClassVar[bool] = ...
    _value2member_map_: ClassVar[dict] = ...
    __and__: ClassVar[Callable] = ...
    __invert__: ClassVar[Callable] = ...
    __or__: ClassVar[Callable] = ...
    __rand__: ClassVar[Callable] = ...
    __ror__: ClassVar[Callable] = ...
    __rxor__: ClassVar[Callable] = ...
    __xor__: ClassVar[Callable] = ...
    def __format__(self, *args, **kwargs) -> str:
        """Convert to a string according to format_spec."""



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

import pytest

import nanoarrow as na
from nanoarrow import device


def test_cpu_device():
    cpu = device.cpu()
    assert cpu.device_type_id == 1
    assert cpu.device_type == device.DeviceType.CPU
    assert cpu.device_id == -1
    assert "device_type: CPU <1>" in repr(cpu)

    cpu2 = device.resolve(1, 0)
    assert cpu2 is cpu


def test_c_device_array():
    # Unrecognized arguments should be passed to c_array() to generate CPU array
    darray = device.c_device_array([1, 2, 3], na.int32())

    assert darray.device_type_id == 1
    assert darray.device_type == device.DeviceType.CPU
    assert darray.device_id == -1
    assert "device_type: CPU <1>" in repr(darray)

    assert darray.schema.format == "i"

    assert len(darray.array) == 3
    assert darray.array.device_type == device.cpu().device_type
    assert darray.array.device_id == device.cpu().device_id

    darray_view = darray.view()
    assert len(darray_view) == 3
    assert list(darray_view.buffer(1)) == [1, 2, 3]

    # A CDeviceArray should be returned as is
    assert device.c_device_array(darray) is darray

    # A CPU device array should be able to export to a regular array
    array = na.c_array(darray)
    assert array.schema.format == "i"
    assert array.buffers == darray.array.buffers


def test_c_device_array_from_dlpack(cuda_device):
    from nanoarrow.device import DeviceType

    cp = pytest.importorskip("cupy")
    if not cuda_device:
        pytest.skip("CUDA device not available")

    assert cuda_device.device_type == DeviceType.CUDA
    assert cuda_device.device_id == 0

    gpu_array = cp.array([1, 2, 3])
    darray = device.c_device_array(gpu_array)
    assert isinstance(darray, device.CDeviceArray)
    assert darray.device_type == DeviceType.CUDA
    assert len(darray.array) == 3
    assert darray.array.null_count == 0

    gpu_buffer_view = darray.array.view().buffer(1)
    cp.testing.assert_array_equal(cp.from_dlpack(gpu_buffer_view), gpu_array)


def test_c_device_array_protocol():
    # Wrapper to prevent c_device_array() from returning early when it detects the
    # input is already a CDeviceArray
    class CDeviceArrayWrapper:
        def __init__(self, obj):
            self.obj = obj

        def __arrow_c_device_array__(self, requested_schema=None):
            return self.obj.__arrow_c_device_array__(requested_schema=requested_schema)

    darray = device.c_device_array([1, 2, 3], na.int32())
    wrapper = CDeviceArrayWrapper(darray)

    darray2 = device.c_device_array(wrapper)
    assert darray2.schema.format == "i"
    assert len(darray2.array) == 3
    assert darray2.array.buffers == darray.array.buffers

    with pytest.raises(NotImplementedError):
        device.c_device_array(wrapper, na.int64())



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

import argparse
import pathlib
import re
import subprocess
import sys


# Generate the nanoarrow_c.pxd file used by the Cython extensions
class PxdGenerator:
    def __init__(self):
        self._define_regexes()

    def generate_pxd(self, file_in, file_out):
        file_in_name = pathlib.Path(file_in).name

        # Read the header
        content = None
        with open(file_in, "r") as input:
            content = input.read()

        # Strip comments
        content = self.re_comment.sub("", content)

        # Replace NANOARROW_MAX_FIXED_BUFFERS with its value
        content = self._preprocess_content(content)

        # Find typedefs, types, and function definitions
        typedefs = self._find_typedefs(content)
        types = self._find_types(content)
        func_defs = self._find_func_defs(content)

        # Make corresponding cython definitions
        typedefs_cython = [self._typdef_to_cython(t, "    ") for t in typedefs]
        types_cython = [self._type_to_cython(t, "    ") for t in types]
        func_defs_cython = [self._func_def_to_cython(d, "    ") for d in func_defs]

        # Unindent the header
        header = self.re_newline_plus_indent.sub("\n", self._pxd_header())

        # Write nanoarrow_c.pxd
        with open(file_out, "wb") as output:
            output.write(header.encode("UTF-8"))

            output.write(
                f'\ncdef extern from "{file_in_name}" nogil:\n'.encode("UTF-8")
            )

            # A few things we add in manually
            self._write_defs(output)

            for type in types_cython:
                output.write(type.encode("UTF-8"))
                output.write(b"\n\n")

            for typedef in typedefs_cython:
                output.write(typedef.encode("UTF-8"))
                output.write(b"\n")

            output.write(b"\n")

            for func_def in func_defs_cython:
                output.write(func_def.encode("UTF-8"))
                output.write(b"\n")

    def _preprocess_content(self, content):
        return content

    def _write_defs(self, output):
        pass

    def _define_regexes(self):
        self.re_comment = re.compile(r"\s*//[^\n]*")
        self.re_typedef = re.compile(r"typedef(?P<typedef>[^;]+)")
        self.re_type = re.compile(
            r"(?P<type>struct|union|enum) (?P<name>Arrow[^ ]+) {(?P<body>[^}]*)}"
        )
        self.re_func_def = re.compile(
            r"\n(static inline |NANOARROW_DLL )(?P<const>const )?(struct |enum )?"
            r"(?P<return_type>[A-Za-z0-9_*]+)\s+"
            r"(?P<name>Arrow[A-Za-z0-9]+)\((?P<args>[^\)]*)\);"
        )
        self.re_tagged_type = re.compile(
            r"(?P<type>struct|union|enum) (?P<name>Arrow[A-Za-z]+)"
        )
        self.re_struct_delim = re.compile(r";\s*")
        self.re_enum_delim = re.compile(r",\s*")
        self.re_whitespace = re.compile(r"\s+")
        self.re_newline_plus_indent = re.compile(r"\n +")

    def _strip_comments(self, content):
        return self.re_comment.sub("", content)

    def _find_typedefs(self, content):
        return [m.groupdict() for m in self.re_typedef.finditer(content)]

    def _find_types(self, content):
        return [m.groupdict() for m in self.re_type.finditer(content)]

    def _find_func_defs(self, content):
        return [m.groupdict() for m in self.re_func_def.finditer(content)]

    def _typdef_to_cython(self, t, indent=""):
        typedef = t["typedef"]
        typedef = self.re_tagged_type.sub(r"\2", typedef)
        return f"{indent}ctypedef {typedef}"

    def _type_to_cython(self, t, indent=""):
        type = t["type"]
        name = t["name"]
        body = self.re_tagged_type.sub(r"\2", t["body"].strip())
        if type == "enum":
            items = [item for item in self.re_enum_delim.split(body) if item]
        else:
            items = [item for item in self.re_struct_delim.split(body) if item]

        cython_body = f"\n{indent}    ".join([""] + items)
        return f"{indent}{type} {name}:{cython_body}"

    def _func_def_to_cython(self, d, indent=""):
        return_type = d["return_type"].strip()
        if d["const"]:
            return_type = "const " + return_type
        name = d["name"]
        args = re.sub(r"\s+", " ", d["args"].strip())
        args = self.re_tagged_type.sub(r"\2", args)

        # Cython doesn't do (void)
        if args == "void":
            args = ""

        return f"{indent}{return_type} {name}({args})"

    def _pxd_header(self):
        return """
        # Licensed to the Apache Software Foundation (ASF) under one
        # or more contributor license agreements.  See the NOTICE file
        # distributed with this work for additional information
        # regarding copyright ownership.  The ASF licenses this file
        # to you under the Apache License, Version 2.0 (the
        # "License"); you may not use this file except in compliance
        # with the License.  You may obtain a copy of the License at
        #
        #   http://www.apache.org/licenses/LICENSE-2.0
        #
        # Unless required by applicable law or agreed to in writing,
        # software distributed under the License is distributed on an
        # "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
        # KIND, either express or implied.  See the License for the
        # specific language governing permissions and limitations
        # under the License.

        # cython: language_level = 3

        """


class NanoarrowPxdGenerator(PxdGenerator):
    def _preprocess_content(self, content):
        content = re.sub(r"NANOARROW_MAX_FIXED_BUFFERS", "3", content)
        content = re.sub(r"NANOARROW_BINARY_VIEW_INLINE_SIZE", "12", content)
        content = re.sub(r"NANOARROW_BINARY_VIEW_PREFIX_SIZE", "4", content)
        return content

    def _pxd_header(self):
        return (
            super()._pxd_header()
            + """
    from libc.stdint cimport int8_t, uint8_t, int16_t, uint16_t
    from libc.stdint cimport int32_t, uint32_t, int64_t, uint64_t
    """
        )

    def _write_defs(self, output):
        output.write(b"\n")
        output.write(b"    cdef int NANOARROW_OK\n")
        output.write(b"    cdef int NANOARROW_MAX_FIXED_BUFFERS\n")
        output.write(b"    cdef int ARROW_FLAG_DICTIONARY_ORDERED\n")
        output.write(b"    cdef int ARROW_FLAG_NULLABLE\n")
        output.write(b"    cdef int ARROW_FLAG_MAP_KEYS_SORTED\n")
        output.write(b"\n")


class NanoarrowDevicePxdGenerator(PxdGenerator):
    def _preprocess_content(self, content):
        self.device_names = re.findall("#define (ARROW_DEVICE_[A-Z0-9_]+)", content)
        return super()._preprocess_content(content)

    def _find_typedefs(self, content):
        return []

    def _pxd_header(self):
        return (
            super()._pxd_header()
            + """
    from libc.stdint cimport int32_t, int64_t
    from nanoarrow_c cimport *
    """
        )

    def _write_defs(self, output):
        output.write(b"\n")
        output.write(b"    ctypedef int32_t ArrowDeviceType\n")
        output.write(b"\n")
        for name in self.device_names:
            output.write(f"    cdef ArrowDeviceType {name}\n".encode())
        output.write(b"\n")


def copy_or_generate_nanoarrow_c(target_dir: pathlib.Path):
    vendored_files = [
        "nanoarrow.h",
        "nanoarrow.c",
        "nanoarrow_ipc.h",
        "nanoarrow_ipc.c",
        "nanoarrow_device.h",
        "nanoarrow_device.c",
    ]
    dst = {name: target_dir / name for name in vendored_files}

    this_dir = pathlib.Path(__file__).parent.resolve()
    arrow_proj_dir = this_dir / "subprojects" / "arrow-nanoarrow"

    subprocess.run(
        [
            sys.executable,
            arrow_proj_dir / "ci" / "scripts" / "bundle.py",
            "--symbol-namespace",
            "PythonPkg",
            "--header-namespace",
            "",
            "--source-output-dir",
            target_dir,
            "--include-output-dir",
            target_dir,
            "--with-device",
            "--with-ipc",
        ],
    )

    if not dst["nanoarrow.h"].exists():
        raise ValueError("Attempt to vendor nanoarrow.c/h failed")


# Runs the pxd generator with some information about the file name
def generate_nanoarrow_pxds(target_dir: pathlib.Path):
    NanoarrowPxdGenerator().generate_pxd(
        target_dir / "nanoarrow.h", target_dir / "nanoarrow_c.pxd"
    )
    NanoarrowDevicePxdGenerator().generate_pxd(
        target_dir / "nanoarrow_device.h",
        target_dir / "nanoarrow_device_c.pxd",
    )


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--output-dir", help="Target directory where files should be written"
    )

    args = parser.parse_args()
    target_dir = pathlib.Path(args.output_dir).resolve()

    copy_or_generate_nanoarrow_c(target_dir)
    generate_nanoarrow_pxds(target_dir)



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

import importlib
import re

import nanoarrow as na


def test_version():
    re_py_version = re.compile(r"^[0-9]+\.[0-9]+\.[0-9]+(\.dev[0-9]+)?$")
    assert re_py_version.match(na.__version__) is not None


def test_c_version():
    re_version = re.compile(r"^[0-9]+\.[0-9]+\.[0-9]+(-SNAPSHOT)?$")
    assert re_version.match(na.c_version()) is not None


def test_metadata_version():
    assert importlib.metadata.version("nanoarrow") == na.__version__



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

import pytest

import nanoarrow as na


def test_type_schema_protocol():
    c_schema = na.c_schema(na.Type.INT32)
    assert c_schema.format == "i"


def test_time_unit_create():
    assert na.TimeUnit.create("s") == na.TimeUnit.SECOND
    assert na.TimeUnit.create("ms") == na.TimeUnit.MILLI
    assert na.TimeUnit.create("us") == na.TimeUnit.MICRO
    assert na.TimeUnit.create("ns") == na.TimeUnit.NANO

    assert na.TimeUnit.create(na.TimeUnit.SECOND) == na.TimeUnit.SECOND


def test_schema_create_c_schema():
    schema_obj = na.int32()
    assert schema_obj.type == na.Type.INT32
    assert schema_obj.name == ""
    assert ("some key" in schema_obj.metadata) is False

    schema_obj2 = na.Schema(schema_obj._c_schema)
    assert schema_obj2.type == schema_obj2.type
    assert schema_obj2._c_schema is schema_obj._c_schema

    schema_obj_non_nullable = na.Schema(na.int32(), nullable=False)
    assert schema_obj_non_nullable.nullable is False

    schema_named = na.Schema(na.int32(), name="some_name")
    assert schema_named.name == "some_name"

    schema_metad = na.Schema(na.int32(), metadata={"some key": "some value"})
    assert b"some key" in schema_metad.metadata
    assert schema_metad.metadata[b"some key"] == b"some value"
    assert dict(schema_metad.metadata.items()) == {b"some key": b"some value"}

    with pytest.raises(ValueError):
        na.Schema(schema_obj._c_schema, some_parameter="some_value")


def test_schema_create_no_params():
    schema_obj = na.int32()
    assert schema_obj.type == na.Type.INT32
    assert schema_obj.nullable is True
    assert repr(schema_obj) == "<Schema> int32"

    schema_obj = na.int32(nullable=False)
    assert schema_obj.nullable is False
    assert "non-nullable" in repr(schema_obj)

    schema_obj = na.Schema(na.Type.INT32, name=False)
    assert schema_obj.name is None

    schema_obj = na.Schema(na.Type.INT32, name="not empty")
    assert schema_obj.name == "not empty"
    assert "'not empty': " in repr(schema_obj)

    msg = "params are only supported for obj of class Type"
    with pytest.raises(ValueError, match=msg):
        na.Schema(na.fixed_size_binary(123), byte_width=12)

    with pytest.raises(ValueError, match=r"^Unused parameter"):
        na.Schema(na.Type.INT32, unused_param="unused_value")


def test_schema_simple():
    assert na.null().type == na.Type.NULL
    assert na.bool_().type == na.Type.BOOL
    assert na.int8().type == na.Type.INT8
    assert na.uint8().type == na.Type.UINT8
    assert na.int16().type == na.Type.INT16
    assert na.uint16().type == na.Type.UINT16
    assert na.int32().type == na.Type.INT32
    assert na.uint32().type == na.Type.UINT32
    assert na.int64().type == na.Type.INT64
    assert na.uint64().type == na.Type.UINT64
    assert na.float16().type == na.Type.HALF_FLOAT
    assert na.float32().type == na.Type.FLOAT
    assert na.float64().type == na.Type.DOUBLE
    assert na.string().type == na.Type.STRING
    assert na.large_string().type == na.Type.LARGE_STRING
    assert na.binary().type == na.Type.BINARY
    assert na.large_binary().type == na.Type.LARGE_BINARY
    assert na.date32().type == na.Type.DATE32
    assert na.date64().type == na.Type.DATE64
    assert na.interval_months().type == na.Type.INTERVAL_MONTHS
    assert na.interval_day_time().type == na.Type.INTERVAL_DAY_TIME
    assert na.interval_month_day_nano().type == na.Type.INTERVAL_MONTH_DAY_NANO
    assert na.binary_view().type == na.Type.BINARY_VIEW
    assert na.string_view().type == na.Type.STRING_VIEW


def test_schema_fixed_size_binary():
    schema_obj = na.fixed_size_binary(byte_width=123)
    assert schema_obj.type == na.Type.FIXED_SIZE_BINARY
    assert schema_obj.byte_width == 123


def test_schema_time():
    schema_obj = na.time32(na.TimeUnit.SECOND)
    assert schema_obj.type == na.Type.TIME32
    assert schema_obj.unit == na.TimeUnit.SECOND

    schema_obj = na.time64(na.TimeUnit.MICRO)
    assert schema_obj.type == na.Type.TIME64
    assert schema_obj.unit == na.TimeUnit.MICRO


def test_schema_timestamp():
    schema_obj = na.timestamp(na.TimeUnit.SECOND)
    assert schema_obj.type == na.Type.TIMESTAMP
    assert schema_obj.unit == na.TimeUnit.SECOND
    assert schema_obj.timezone is None

    schema_obj = na.timestamp(na.TimeUnit.SECOND, timezone="America/Halifax")
    assert schema_obj.timezone == "America/Halifax"


def test_schema_duration():
    schema_obj = na.duration(na.TimeUnit.SECOND)
    assert schema_obj.type == na.Type.DURATION
    assert schema_obj.unit == na.TimeUnit.SECOND


def test_schema_decimal():
    schema_obj = na.decimal128(10, 3)
    assert schema_obj.type == na.Type.DECIMAL128
    assert schema_obj.precision == 10
    assert schema_obj.scale == 3

    schema_obj = na.decimal256(10, 3)
    assert schema_obj.type == na.Type.DECIMAL256
    assert schema_obj.precision == 10
    assert schema_obj.scale == 3


def test_schema_struct():
    # Make sure we can use just a list
    schema_obj = na.struct([na.Type.INT32])
    assert schema_obj.type == na.Type.STRUCT
    assert schema_obj.n_fields == 1
    assert schema_obj.field(0).type == na.Type.INT32
    assert schema_obj.field(0).name == ""
    for field in schema_obj.fields:
        assert isinstance(field, na.Schema)

    # Make sure we can use a dictionary to specify fields
    schema_obj = na.struct({"col_name": na.Type.INT32})
    assert schema_obj.type == na.Type.STRUCT
    assert schema_obj.field(0).type == na.Type.INT32
    assert schema_obj.field(0).name == "col_name"

    # Make sure we can use a Schema when constructing fields (and that
    # field names are taken from the input)
    schema_obj = na.struct([schema_obj.field(0)])
    assert schema_obj.type == na.Type.STRUCT
    assert schema_obj.field(0).type == na.Type.INT32
    assert schema_obj.field(0).name == "col_name"


def test_schema_list_():
    schema_obj = na.list_(na.null())
    assert schema_obj.type == na.Type.LIST
    assert schema_obj.value_type.type == na.Type.NULL


def test_schema_large_list():
    schema_obj = na.large_list(na.null())
    assert schema_obj.type == na.Type.LARGE_LIST
    assert schema_obj.value_type.type == na.Type.NULL


def test_schema_fixed_size_list():
    schema_obj = na.fixed_size_list(na.null(), 123)
    assert schema_obj.type == na.Type.FIXED_SIZE_LIST
    assert schema_obj.value_type.type == na.Type.NULL
    assert schema_obj.list_size == 123


def test_schema_map():
    schema_obj = na.map_(na.int32(), na.string())
    assert schema_obj.type == na.Type.MAP
    assert schema_obj.key_type.type == na.Type.INT32
    assert schema_obj.value_type.type == na.Type.STRING
    assert schema_obj.keys_sorted is False

    assert na.map_(na.int32(), na.string(), keys_sorted=True).keys_sorted is True


def test_schema_dictionary():
    schema_obj = na.dictionary(na.int8(), na.null())
    assert schema_obj.type == na.Type.DICTIONARY
    assert schema_obj.index_type.type == na.Type.INT8
    assert schema_obj.value_type.type == na.Type.NULL
    assert schema_obj.dictionary_ordered is False

    schema_obj_ordered = na.dictionary(na.int8(), na.null(), dictionary_ordered=True)
    assert schema_obj_ordered.type == na.Type.DICTIONARY
    assert schema_obj_ordered.index_type.type == na.Type.INT8
    assert schema_obj_ordered.value_type.type == na.Type.NULL
    assert schema_obj_ordered.dictionary_ordered is True


def test_schema_union():
    schema_obj = na.dense_union([na.int32(), na.string()])
    assert schema_obj.type == na.Type.DENSE_UNION
    assert schema_obj.type_codes == [0, 1]

    # Make sure custom type_codes come through
    schema_obj = na.dense_union([na.int32(), na.string()], type_codes=[1, 0])
    assert schema_obj.type == na.Type.DENSE_UNION
    assert schema_obj.type_codes == [1, 0]

    schema_obj = na.sparse_union([na.int32(), na.string()])
    assert schema_obj.type == na.Type.SPARSE_UNION
    assert schema_obj.type_codes == [0, 1]

    # Make sure custom type_codes come through
    schema_obj = na.sparse_union([na.int32(), na.string()], type_codes=[1, 0])
    assert schema_obj.type == na.Type.SPARSE_UNION
    assert schema_obj.type_codes == [1, 0]


def test_schema_extension():
    schema_obj = na.int32()
    assert schema_obj.extension is None

    # Check with metadata manually added
    schema_obj = na.Schema(
        na.int32(),
        metadata={
            "ARROW:extension:name": "arrow.test",
            "ARROW:extension:metadata": "abcdefg",
        },
    )
    assert schema_obj.extension.name == "arrow.test"
    assert schema_obj.extension.metadata == b"abcdefg"

    # Check from extension_type constructor
    schema_obj = na.extension_type(na.int32(), "arrow.test", "abcdefg")
    assert schema_obj.type == na.Type.EXTENSION
    assert schema_obj.extension is not None
    assert schema_obj.extension.name == "arrow.test"
    assert schema_obj.extension.metadata == b"abcdefg"
    assert schema_obj.extension.storage.type == na.Type.INT32
    assert schema_obj.nullable is True

    schema_obj = na.extension_type(na.int32(), "arrow.test", nullable=False)
    assert schema_obj.extension.name == "arrow.test"
    assert schema_obj.extension.metadata == b""
    assert schema_obj.extension.storage.type == na.Type.INT32
    assert schema_obj.nullable is False


def test_schema_alias_constructor():
    schema = na.schema(na.Type.INT32)
    assert isinstance(schema, na.Schema)


def test_schema_serialize():
    import io

    schema = na.struct({"some_col": na.int32()}, nullable=False)

    serialized = schema.serialize()
    schema_roundtrip = na.ArrayStream.from_readable(serialized).schema
    assert repr(schema_roundtrip) == repr(schema)

    out = io.BytesIO()
    schema.serialize(out)
    schema_roundtrip = na.ArrayStream.from_readable(out.getvalue()).schema
    assert repr(schema_roundtrip) == repr(schema)


def test_schema_repr():
    schema = na.struct(
        {
            "col1": na.int32(),
            "col2": na.int16(),
            "col3": na.string(),
            "col4": na.timestamp(unit=na.TimeUnit.SECOND),
        },
        nullable=False,
    )

    assert repr(schema) == (
        "<Schema> non-nullable struct"
        "<col1: int32, col2: int16, col3: string, "
        "col4: timestamp('s', '')>"
    )



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

# cython: language_level = 3

from libc.stdint cimport uintptr_t, uint8_t, int64_t
from cpython.pycapsule cimport PyCapsule_GetPointer
from cpython.unicode cimport PyUnicode_AsUTF8AndSize
from cpython cimport (
    Py_buffer,
    PyBuffer_Release,
    PyBUF_ANY_CONTIGUOUS,
    PyBUF_FORMAT,
    PyBytes_FromStringAndSize,
    PyObject_GetBuffer,
    PyUnicode_FromStringAndSize,
)

from nanoarrow_c cimport (
    ArrowArray,
    ArrowArrayAddVariadicBuffers,
    ArrowArrayAppendBytes,
    ArrowArrayAppendNull,
    ArrowArrayAppendString,
    ArrowArrayBuffer,
    ArrowArrayFinishBuilding,
    ArrowArrayInitFromSchema,
    ArrowArrayInitFromType,
    ArrowArrayMove,
    ArrowArrayRelease,
    ArrowArrayStartAppending,
    ArrowArrayView,
    ArrowArrayViewComputeNullCount,
    ArrowArrayViewInitFromSchema,
    ArrowArrayViewIsNull,
    ArrowArrayViewGetBytesUnsafe,
    ArrowArrayViewGetBufferDataType,
    ArrowArrayViewGetBufferElementSizeBits,
    ArrowArrayViewGetBufferType,
    ArrowArrayViewGetBufferView,
    ArrowArrayViewGetNumBuffers,
    ArrowArrayViewGetStringUnsafe,
    ArrowArrayViewSetArray,
    ArrowArrayViewSetArrayMinimal,
    ArrowBitCountSet,
    ArrowBuffer,
    ArrowBufferMove,
    ArrowBufferType,
    ArrowBufferView,
    ArrowSchemaInitFromType,
    ArrowStringView,
    ArrowType,
    ArrowTypeString,
    ArrowValidationLevel,
    NANOARROW_BUFFER_TYPE_DATA,
    NANOARROW_BUFFER_TYPE_DATA_OFFSET,
    NANOARROW_BUFFER_TYPE_VARIADIC_DATA,
    NANOARROW_BUFFER_TYPE_VARIADIC_SIZE,
    NANOARROW_BUFFER_TYPE_TYPE_ID,
    NANOARROW_BUFFER_TYPE_UNION_OFFSET,
    NANOARROW_BUFFER_TYPE_VALIDITY,
    NANOARROW_VALIDATION_LEVEL_DEFAULT,
    NANOARROW_VALIDATION_LEVEL_FULL,
    NANOARROW_VALIDATION_LEVEL_MINIMAL,
    NANOARROW_VALIDATION_LEVEL_NONE,
    NANOARROW_OK,
)

from nanoarrow_device_c cimport (
    ARROW_DEVICE_CPU,
    ArrowDeviceType,
    ArrowDeviceArray,
    ArrowDeviceArrayInit,
)

from nanoarrow._device cimport Device, CSharedSyncEvent

from nanoarrow._buffer cimport CBuffer, CBufferView
from nanoarrow._schema cimport CSchema, CLayout
from nanoarrow._utils cimport (
    alloc_c_array,
    alloc_c_device_array,
    alloc_c_array_view,
    c_array_shallow_copy,
    c_device_array_shallow_copy,
    Error
)

from typing import Iterable, Tuple, Union

from nanoarrow import _repr_utils
from nanoarrow._device import DEVICE_CPU, DeviceType


cdef class CArrayView:
    """Low-level ArrowArrayView wrapper

    This object is a literal wrapper around an ArrowArrayView. It provides field accessors
    that return Python objects and handles the structure lifecycle (i.e., initialized
    ArrowArrayView structures are always released).

    See `nanoarrow.c_array_view()` for construction and usage examples.
    """

    def __cinit__(self, object base, uintptr_t addr):
        self._base = base
        self._ptr = <ArrowArrayView*>addr
        self._event = CSharedSyncEvent(DEVICE_CPU)

    def _set_array(self, CArray array, Device device=DEVICE_CPU):
        cdef Error error = Error()
        cdef int code

        if device is DEVICE_CPU:
            code = ArrowArrayViewSetArray(self._ptr, array._ptr, &error.c_error)
        else:
            code = ArrowArrayViewSetArrayMinimal(self._ptr, array._ptr, &error.c_error)

        error.raise_message_not_ok("ArrowArrayViewSetArray()", code)
        self._array_base = array._base
        self._event = CSharedSyncEvent(device, <uintptr_t>array._sync_event)

        return self

    @property
    def storage_type_id(self):
        return self._ptr.storage_type

    @property
    def storage_type(self):
        cdef const char* type_str = ArrowTypeString(self._ptr.storage_type)
        if type_str != NULL:
            return type_str.decode('UTF-8')

    @property
    def layout(self):
        return CLayout(self, <uintptr_t>&self._ptr.layout)

    def __len__(self):
        return self._ptr.length

    @property
    def length(self):
        return len(self)

    @property
    def offset(self):
        return self._ptr.offset

    @property
    def null_count(self):
        if self._ptr.null_count != -1:
            return self._ptr.null_count

        cdef ArrowBufferType buffer_type = self._ptr.layout.buffer_type[0]
        cdef const uint8_t* validity_bits = self._ptr.buffer_views[0].data.as_uint8

        if buffer_type != NANOARROW_BUFFER_TYPE_VALIDITY:
            self._ptr.null_count = 0
        elif validity_bits == NULL:
            self._ptr.null_count = 0
        elif self._event.device is DEVICE_CPU:
            self._ptr.null_count = ArrowArrayViewComputeNullCount(self._ptr)

        return self._ptr.null_count

    @property
    def n_children(self):
        return self._ptr.n_children

    def child(self, int64_t i):
        if i < 0 or i >= self._ptr.n_children:
            raise IndexError(f"{i} out of range [0, {self._ptr.n_children})")

        cdef CArrayView child = CArrayView(
            self._base,
            <uintptr_t>self._ptr.children[i]
        )

        child._event = self._event

        return child

    @property
    def children(self):
        for i in range(self.n_children):
            yield self.child(i)

    @property
    def n_buffers(self):
        return ArrowArrayViewGetNumBuffers(self._ptr)

    def _buffer_info(self, int64_t i):
        if i < 0 or i >= self.n_buffers:
            raise IndexError(f"{i} out of range [0, {self.n_buffers}]")

        cdef ArrowBufferView view = ArrowArrayViewGetBufferView(self._ptr, i)

        return (
            ArrowArrayViewGetBufferType(self._ptr, i),
            ArrowArrayViewGetBufferDataType(self._ptr, i),
            ArrowArrayViewGetBufferElementSizeBits(self._ptr, i),
            <uintptr_t>view.data.data,
            view.size_bytes
        )

    def buffer_type(self, int64_t i):
        buffer_type = self._buffer_info(i)[0]
        if buffer_type == NANOARROW_BUFFER_TYPE_VALIDITY:
            return "validity"
        elif buffer_type == NANOARROW_BUFFER_TYPE_TYPE_ID:
            return "type_id"
        elif buffer_type == NANOARROW_BUFFER_TYPE_UNION_OFFSET:
            return "union_offset"
        elif buffer_type == NANOARROW_BUFFER_TYPE_DATA_OFFSET:
            return "data_offset"
        elif buffer_type == NANOARROW_BUFFER_TYPE_DATA:
            return "data"
        elif buffer_type == NANOARROW_BUFFER_TYPE_VARIADIC_DATA:
            return "variadic_data"
        elif buffer_type == NANOARROW_BUFFER_TYPE_VARIADIC_SIZE:
            return "variadic_size"
        else:
            return "none"

    def buffer(self, int64_t i):
        _, data_type, element_size_bits, addr, size = self._buffer_info(i)

        cdef ArrowBufferView buffer_view
        buffer_view.data.data = <void*>addr
        buffer_view.size_bytes = size

        # Check the buffer size here because the error later is cryptic.
        # Buffer sizes are set to -1 when they are "unknown", so because of errors
        # in nanoarrow/C or because the array is on a non-CPU device, that -1 value
        # could leak its way here.
        if buffer_view.size_bytes < 0:
            raise RuntimeError(f"ArrowArrayView buffer {i} has size_bytes < 0")

        return CBufferView(
            self._array_base,
            addr,
            size,
            data_type,
            element_size_bits,
            self._event
        )

    @property
    def buffers(self):
        for i in range(self.n_buffers):
            yield self.buffer(i)

    @property
    def dictionary(self):
        if self._ptr.dictionary == NULL:
            return None

        cdef CArrayView dictionary = CArrayView(
            self,
            <uintptr_t>self._ptr.dictionary
        )
        dictionary._event = self._event

        return dictionary

    def _iter_bytes(self, int64_t offset, int64_t length) -> bytes | None:
        cdef ArrowBufferView item_view
        for i in range(offset, length):
            if ArrowArrayViewIsNull(self._ptr, i):
                yield None
            else:
                item_view = ArrowArrayViewGetBytesUnsafe(self._ptr, i)
                yield PyBytes_FromStringAndSize(item_view.data.as_char, item_view.size_bytes)

    def _iter_str(self, int64_t offset, int64_t length) -> str | None:
        cdef ArrowStringView item_view
        for i in range(offset, length):
            if ArrowArrayViewIsNull(self._ptr, i):
                yield None
            else:
                item_view = ArrowArrayViewGetStringUnsafe(self._ptr, i)
                yield PyUnicode_FromStringAndSize(item_view.data, item_view.size_bytes)

    def __repr__(self):
        return _repr_utils.array_view_repr(self)

    @staticmethod
    def from_schema(CSchema schema):
        cdef ArrowArrayView* c_array_view
        base = alloc_c_array_view(&c_array_view)

        cdef Error error = Error()
        cdef int code = ArrowArrayViewInitFromSchema(c_array_view,
                                                     schema._ptr, &error.c_error)
        error.raise_message_not_ok("ArrowArrayViewInitFromSchema()", code)

        return CArrayView(base, <uintptr_t>c_array_view)

    @staticmethod
    def from_array(CArray array, Device device=DEVICE_CPU):
        out = CArrayView.from_schema(array._schema)
        return out._set_array(array, device)


cdef class CArray:
    """Low-level ArrowArray wrapper

    This object is a literal wrapper around a read-only ArrowArray. It provides field accessors
    that return Python objects and handles the C Data interface lifecycle (i.e., initialized
    ArrowArray structures are always released).

    See `nanoarrow.c_array()` for construction and usage examples.
    """

    @staticmethod
    def allocate(CSchema schema) -> CArray:
        """Allocate a released ArrowArray"""
        cdef ArrowArray* c_array_out
        base = alloc_c_array(&c_array_out)
        return CArray(base, <uintptr_t>c_array_out, schema)

    def __cinit__(self, object base, uintptr_t addr, CSchema schema):
        self._base = base
        self._ptr = <ArrowArray*>addr
        self._schema = schema
        self._device_type = ARROW_DEVICE_CPU
        self._device_id = -1
        self._sync_event = NULL

    cdef _set_device(self, ArrowDeviceType device_type, int64_t device_id, void* sync_event):
        self._device_type = device_type
        self._device_id = device_id
        self._sync_event = sync_event

    @staticmethod
    def _import_from_c_capsule(schema_capsule, array_capsule) -> CArray:
        """Import from a ArrowSchema and ArrowArray PyCapsule tuple.

        Parameters
        ----------
        schema_capsule : PyCapsule
            A valid PyCapsule with name 'arrow_schema' containing an
            ArrowSchema pointer.
        array_capsule : PyCapsule
            A valid PyCapsule with name 'arrow_array' containing an
            ArrowArray pointer.
        """
        cdef:
            CSchema out_schema
            CArray out

        out_schema = CSchema._import_from_c_capsule(schema_capsule)
        out = CArray(
            array_capsule,
            <uintptr_t>PyCapsule_GetPointer(array_capsule, 'arrow_array'),
            out_schema
        )

        return out

    def __getitem__(self, k) -> CArray:
        self._assert_valid()

        if not isinstance(k, slice):
            raise TypeError(
                f"Can't subset CArray with object of type {type(k).__name__}")

        if k.step is not None:
            raise ValueError("Can't slice CArray with step")

        cdef int64_t start = 0 if k.start is None else k.start
        cdef int64_t stop = self._ptr.length if k.stop is None else k.stop
        if start < 0:
            start = self._ptr.length + start
        if stop < 0:
            stop = self._ptr.length + stop

        if start > self._ptr.length or stop > self._ptr.length or stop < start:
            raise IndexError(
                f"{k} does not describe a valid slice of CArray "
                f"with length {self._ptr.length}"
            )

        cdef ArrowArray* c_array_out
        base = alloc_c_array(&c_array_out)
        c_array_shallow_copy(self._base, self._ptr, c_array_out)

        c_array_out.offset = c_array_out.offset + start
        c_array_out.length = stop - start
        cdef CArray out = CArray(base, <uintptr_t>c_array_out, self._schema)
        out._set_device(self._device_type, self._device_id, self._sync_event)

        return out

    def __arrow_c_array__(self, requested_schema=None):
        """
        Get a pair of PyCapsules containing a C ArrowArray representation of the object.

        Parameters
        ----------
        requested_schema : PyCapsule | None
            A PyCapsule containing a C ArrowSchema representation of a requested
            schema. Not supported.

        Returns
        -------
        Tuple[PyCapsule, PyCapsule]
            A pair of PyCapsules containing a C ArrowSchema and ArrowArray,
            respectively.
        """
        self._assert_valid()

        if self._device_type != ARROW_DEVICE_CPU:
            raise ValueError(
                "Can't invoke __arrow_c_array__ on non-CPU array "
                f"with device_type {self._device_type}")

        if requested_schema is not None:
            raise NotImplementedError("requested_schema")

        # Export a shallow copy pointing to the same data in a way
        # that ensures this object stays valid.

        # TODO optimize this to export a version where children are reference
        # counted and can be released separately
        cdef ArrowArray* c_array_out
        array_capsule = alloc_c_array(&c_array_out)
        c_array_shallow_copy(self._base, self._ptr, c_array_out)

        return self._schema.__arrow_c_schema__(), array_capsule

    def _addr(self) -> int:
        return <uintptr_t>self._ptr

    def is_valid(self) -> bool:
        """Check for a non-null and non-released underlying ArrowArray"""
        return self._ptr != NULL and self._ptr.release != NULL

    def _assert_valid(self):
        if self._ptr == NULL:
            raise RuntimeError("CArray is NULL")
        if self._ptr.release == NULL:
            raise RuntimeError("CArray is released")

    def view(self) -> CArrayView:
        """Allocate a :class:`CArrayView` to access the buffers of this array"""
        device = Device.resolve(self._device_type, self._device_id)
        return CArrayView.from_array(self, device)

    @property
    def schema(self) -> CSchema:
        return self._schema

    @property
    def device_type(self) -> DeviceType:
        return DeviceType(self._device_type)

    @property
    def device_type_id(self) -> int:
        return self._device_type

    @property
    def device_id(self) -> int:
        return self._device_id

    def __len__(self) -> int:
        self._assert_valid()
        return self._ptr.length

    @property
    def length(self) -> int:
        return len(self)

    @property
    def offset(self) -> int:
        self._assert_valid()
        return self._ptr.offset

    @property
    def null_count(self) -> int:
        self._assert_valid()
        return self._ptr.null_count

    @property
    def n_buffers(self) -> int:
        self._assert_valid()
        return self._ptr.n_buffers

    @property
    def buffers(self) -> Tuple[int, ...]:
        self._assert_valid()
        return tuple(<uintptr_t>self._ptr.buffers[i] for i in range(self._ptr.n_buffers))

    @property
    def n_children(self) -> int:
        self._assert_valid()
        return self._ptr.n_children

    def child(self, int64_t i):
        self._assert_valid()
        if i < 0 or i >= self._ptr.n_children:
            raise IndexError(f"{i} out of range [0, {self._ptr.n_children})")
        cdef CArray out = CArray(
            self._base,
            <uintptr_t>self._ptr.children[i],
            self._schema.child(i)
        )
        out._set_device(self._device_type, self._device_id, self._sync_event)
        return out

    @property
    def children(self) -> Iterable[CArray]:
        for i in range(self.n_children):
            yield self.child(i)

    @property
    def dictionary(self) -> Union[CArray, None]:
        self._assert_valid()
        cdef CArray out
        if self._ptr.dictionary != NULL:
            out = CArray(self, <uintptr_t>self._ptr.dictionary, self._schema.dictionary)
            out._set_device(self._device_type, self._device_id, self._sync_event)
            return out
        else:
            return None

    def __repr__(self) -> str:
        return _repr_utils.array_repr(self)


cdef class CArrayBuilder:
    """Helper for constructing an ArrowArray

    The primary function of this class is to wrap the nanoarrow C library calls
    that build up the components of an ArrowArray.
    """
    cdef CArray c_array
    cdef ArrowArray* _ptr
    cdef Device _device
    cdef bint _can_validate

    def __cinit__(self, CArray array, Device device=DEVICE_CPU):
        self.c_array = array
        self._ptr = array._ptr
        self._device = device
        self._can_validate = device is DEVICE_CPU

    @staticmethod
    def allocate(Device device=DEVICE_CPU):
        """Create a CArrayBuilder

        Allocates memory for an ArrowArray and populates it with nanoarrow's
        ArrowArray private_data/release callback implementation. This should
        usually be followed by :meth:`init_from_type` or :meth:`init_from_schema`.
        """
        return CArrayBuilder(CArray.allocate(CSchema.allocate()), device)

    def is_empty(self) -> bool:
        """Check if any items have been appended to this builder"""
        if self._ptr.release == NULL:
            raise RuntimeError("CArrayBuilder is not initialized")

        return self._ptr.length == 0

    def init_from_type(self, int type_id) -> CArrayBuilder:
        if self._ptr.release != NULL:
            raise RuntimeError("CArrayBuilder is already initialized")

        cdef int code = ArrowArrayInitFromType(self._ptr, <ArrowType>type_id)
        Error.raise_error_not_ok("ArrowArrayInitFromType()", code)

        code = ArrowSchemaInitFromType(self.c_array._schema._ptr, <ArrowType>type_id)
        Error.raise_error_not_ok("ArrowSchemaInitFromType()", code)

        return self

    def init_from_schema(self, CSchema schema) -> CArrayBuilder:
        if self._ptr.release != NULL:
            raise RuntimeError("CArrayBuilder is already initialized")

        cdef Error error = Error()
        cdef int code = ArrowArrayInitFromSchema(self._ptr, schema._ptr, &error.c_error)
        error.raise_message_not_ok("ArrowArrayInitFromType()", code)

        self.c_array._schema = schema
        return self

    def start_appending(self) -> CArrayBuilder:
        """Use append mode for building this ArrowArray

        Calling this method is required to produce a valid array prior to calling
        :meth:`append_strings` or `append_bytes`.
        """
        if self._device != DEVICE_CPU:
            raise ValueError("Can't append to non-CPU array")

        cdef int code = ArrowArrayStartAppending(self._ptr)
        Error.raise_error_not_ok("ArrowArrayStartAppending()", code)
        return self

    def append_strings(self, obj: Iterable[Union[str, None]]) -> CArrayBuilder:
        cdef int code
        cdef Py_ssize_t item_utf8_size
        cdef ArrowStringView item

        for py_item in obj:
            if py_item is None:
                code = ArrowArrayAppendNull(self._ptr, 1)
            else:
                # Cython raises the error from PyUnicode_AsUTF8AndSize()
                # in the event that py_item is not a str(); however, we
                # set item_utf8_size = 0 to be safe.
                item_utf8_size = 0
                item.data = PyUnicode_AsUTF8AndSize(py_item, &item_utf8_size)
                item.size_bytes = item_utf8_size
                code = ArrowArrayAppendString(self._ptr, item)

            if code != NANOARROW_OK:
                Error.raise_error(f"append string item {py_item}", code)

        return self

    def append_bytes(self, obj: Iterable[Union[str, None]]) -> CArrayBuilder:
        cdef Py_buffer buffer
        cdef ArrowBufferView item

        for py_item in obj:
            if py_item is None:
                code = ArrowArrayAppendNull(self._ptr, 1)
            else:
                PyObject_GetBuffer(py_item, &buffer, PyBUF_ANY_CONTIGUOUS | PyBUF_FORMAT)

                if buffer.ndim != 1:
                    raise ValueError("Can't append buffer with dimensions != 1 to binary array")

                if buffer.itemsize != 1:
                    PyBuffer_Release(&buffer)
                    raise ValueError("Can't append buffer with itemsize != 1 to binary array")

                item.data.data = buffer.buf
                item.size_bytes = buffer.len
                code = ArrowArrayAppendBytes(self._ptr, item)
                PyBuffer_Release(&buffer)

            if code != NANOARROW_OK:
                Error.raise_error(f"append bytes item {py_item}", code)

    def set_offset(self, int64_t offset) -> CArrayBuilder:
        self.c_array._assert_valid()
        self._ptr.offset = offset
        return self

    def set_length(self, int64_t length) -> CArrayBuilder:
        self.c_array._assert_valid()
        self._ptr.length = length
        return self

    def set_null_count(self, int64_t null_count) -> CArrayBuilder:
        self.c_array._assert_valid()
        self._ptr.null_count = null_count
        return self

    def resolve_null_count(self) -> CArrayBuilder:
        """Ensure the output null count is synchronized with existing buffers

        Note that this will not attempt to access non-CPU buffers such that
        :attr:`null_count` might still be -1 after calling this method.
        """
        self.c_array._assert_valid()

        # This doesn't apply to unions. We currently don't have a schema view
        # or array view we can use to query the type ID, so just use the format
        # string for now.
        format = self.c_array.schema.format
        if format.startswith("+us:") or format.startswith("+ud:"):
            return self

        # Don't overwrite an explicit null count
        if self._ptr.null_count != -1:
            return self

        cdef ArrowBuffer* validity_buffer = ArrowArrayBuffer(self._ptr, 0)
        if validity_buffer.size_bytes == 0:
            self._ptr.null_count = 0
            return self

        # Don't attempt to access a non-cpu buffer
        if self._device != DEVICE_CPU:
            return self

        # From _ArrowBytesForBits(), which is not included in nanoarrow_c.pxd
        # because it's an internal inline function.
        cdef int64_t bits = self._ptr.offset + self._ptr.length
        cdef int64_t bytes_required = (bits >> 3) + ((bits & 7) != 0)

        if validity_buffer.size_bytes < bytes_required:
            raise ValueError(
                f"Expected validity bitmap >= {bytes_required} bytes "
                f"but got validity bitmap with {validity_buffer.size_bytes} bytes"
            )

        cdef int64_t count = ArrowBitCountSet(
            validity_buffer.data,
            self._ptr.offset,
            self._ptr.length
        )
        self._ptr.null_count = self._ptr.length - count
        return self

    def ensure_buffers(self, n_buffers) -> CArrayBuilder:
        if self._ptr.n_buffers < n_buffers:
            code = ArrowArrayAddVariadicBuffers(self._ptr, n_buffers - self._ptr.n_buffers)
            Error.raise_error_not_ok("ArrowArrayAddVariadicBuffers() failed", code)

        return self

    def set_buffer(self, int64_t i, CBuffer buffer, move=False) -> CArrayBuilder:
        """Set an ArrowArray buffer

        Sets a buffer of this ArrowArray such the pointer at array->buffers[i] is
        equal to buffer->data and such that the buffer's lifcycle is managed by
        the array. If move is True, the input Python object that previously wrapped
        the ArrowBuffer will be invalidated, which is usually the desired behaviour
        if you built or imported a buffer specifically to build this array. If move
        is False (the default), this function will a make a shallow copy via another
        layer of Python object wrapping.
        """
        if i < 0 or i >= self._ptr.n_buffers:
            raise IndexError(f"i must be >= 0 and < {self._ptr.n_buffers}")

        if buffer._device != self._device:
            raise ValueError(
                f"Builder device ({self._device.device_type}/{self._device.device_id})"
                " and buffer device "
                f"({buffer._device.device_type}/{buffer._device.device_id})"
                " are not identical"
            )

        self.c_array._assert_valid()
        if not move:
            buffer = CBuffer.from_pybuffer(buffer)

        ArrowBufferMove(buffer._ptr, ArrowArrayBuffer(self._ptr, i))

        return self

    def set_child(self, int64_t i, CArray c_array, move=False) -> CArrayBuilder:
        """Set an ArrowArray child

        Set a child of this array by performing a show copy or optionally
        transferring ownership to this object. The initialized child array
        must have been initialized before this call by initializing this
        builder with a schema containing the correct number of children.
        """
        cdef CArray child = self.c_array.child(i)
        if child._ptr.release != NULL:
            ArrowArrayRelease(child._ptr)

        if (
            self._device.device_type_id != c_array.device_type_id
            or self._device.device_id != c_array.device_id
        ):
            raise ValueError(
                f"Builder device ({self._device.device_type}/{self._device.device_id})"
                " and child device "
                f"({c_array.device_type}/{c_array.device_id}) are not identical"
            )

        # There is probably a way to avoid a full synchronize for each child
        # (e.g., perhaps the ArrayBuilder could allocate a stream to use such
        # that an event can be allocated on finish_device() and synchronization
        # could be avoided entirely). Including this for now for safety.
        cdef CSharedSyncEvent sync = CSharedSyncEvent(
            self._device,
            <uintptr_t>c_array._sync_event
        )
        sync.synchronize()

        if not move:
            c_array_shallow_copy(c_array._base, c_array._ptr, child._ptr)
        else:
            ArrowArrayMove(c_array._ptr, child._ptr)

        # After setting children, we can't use the built-in validation done by
        # ArrowArrayFinishBuilding() because it assumes that the private_data of
        # each array (recursively) is one that was initialized by ArrowArrayInit()
        self._can_validate = False

        return self

    def finish(self, validation_level=None) -> CArray:
        """Finish building this array

        Performs any steps required to return a valid ArrowArray and optionally
        validates the output to ensure that the result is valid (given the information
        the array has available to it).

        Parameters
        ----------
        validation_level : None, "full", "default", "minimal", or "none", optional
            Explicitly define a validation level or use None to perform default
            validation if possible. Validation may not be possible if children
            were set that were not created by nanoarrow.
        """
        self.c_array._assert_valid()
        cdef ArrowValidationLevel c_validation_level
        cdef Error error = Error()
        cdef int code

        if not self._can_validate and validation_level not in (None, "none"):
            raise NotImplementedError("Validation for array with children is not implemented")

        if not self._can_validate:
            validation_level = "none"

        c_validation_level = NANOARROW_VALIDATION_LEVEL_DEFAULT
        if validation_level == "full":
            c_validation_level = NANOARROW_VALIDATION_LEVEL_FULL
        elif validation_level == "minimal":
            c_validation_level = NANOARROW_VALIDATION_LEVEL_MINIMAL
        elif validation_level == "none":
            c_validation_level = NANOARROW_VALIDATION_LEVEL_NONE

        code = ArrowArrayFinishBuilding(self._ptr, c_validation_level, &error.c_error)
        error.raise_message_not_ok("ArrowArrayFinishBuildingDefault()", code)

        out = self.c_array
        self.c_array = CArray.allocate(CSchema.allocate())
        self._ptr = self.c_array._ptr
        self._can_validate = True

        return out

    def finish_device(self):
        """Finish building this array and export to an ArrowDeviceArray

        Calls :meth:`finish`, propagating device information into an ArrowDeviceArray.
        """
        cdef CArray array = self.finish()

        cdef ArrowDeviceArray* device_array_ptr
        holder = alloc_c_device_array(&device_array_ptr)
        cdef int code = ArrowDeviceArrayInit(self._device._ptr, device_array_ptr, array._ptr, NULL)
        Error.raise_error_not_ok("ArrowDeviceArrayInit", code)

        return CDeviceArray(holder, <uintptr_t>device_array_ptr, array._schema)


cdef class CDeviceArray:
    """Low-level ArrowDeviceArray wrapper

    This object is a literal wrapper around an ArrowDeviceArray. It provides field accessors
    that return Python objects and handles the structure lifecycle (i.e., initialized
    ArrowDeviceArray structures are always released).

    See `nanoarrow.device.c_device_array()` for construction and usage examples.
    """

    def __cinit__(self, object base, uintptr_t addr, CSchema schema):
        self._base = base
        self._ptr = <ArrowDeviceArray*>addr
        self._schema = schema

    @staticmethod
    def _init_from_array(Device device, uintptr_t array_addr, CSchema schema):
        cdef ArrowArray* array_ptr = <ArrowArray*>array_addr
        cdef ArrowDeviceArray* device_array_ptr
        cdef void* sync_event = NULL
        holder = alloc_c_device_array(&device_array_ptr)
        cdef int code = ArrowDeviceArrayInit(device._ptr, device_array_ptr, array_ptr, sync_event)
        Error.raise_error_not_ok("ArrowDeviceArrayInit", code)

        return CDeviceArray(holder, <uintptr_t>device_array_ptr, schema)

    @property
    def schema(self) -> CSchema:
        return self._schema

    @property
    def device_type(self) -> DeviceType:
        return DeviceType(self._ptr.device_type)

    @property
    def device_type_id(self) -> int:
        return self._ptr.device_type

    @property
    def device_id(self) -> int:
        return self._ptr.device_id

    @property
    def array(self) -> CArray:
        cdef CArray array = CArray(self, <uintptr_t>&self._ptr.array, self._schema)
        array._set_device(self._ptr.device_type, self._ptr.device_id, self._ptr.sync_event)
        return array

    def view(self) -> CArrayView:
        return self.array.view()

    def __arrow_c_array__(self, requested_schema=None):
        return self.array.__arrow_c_array__(requested_schema=requested_schema)

    def __arrow_c_device_array__(self, requested_schema=None):
        if requested_schema is not None:
            raise NotImplementedError("requested_schema")

        # TODO: evaluate whether we need to synchronize here or whether we should
        # move device arrays instead of shallow-copying them
        cdef ArrowDeviceArray* c_array_out
        device_array_capsule = alloc_c_device_array(&c_array_out)
        c_device_array_shallow_copy(self._base, self._ptr, c_array_out)

        return self._schema.__arrow_c_schema__(), device_array_capsule

    @staticmethod
    def _import_from_c_capsule(schema_capsule, device_array_capsule) -> CDeviceArray:
        """
        Import from an ArrowSchema and ArrowArray PyCapsule tuple.

        Parameters
        ----------
        schema_capsule : PyCapsule
            A valid PyCapsule with name 'arrow_schema' containing an
            ArrowSchema pointer.
        device_array_capsule : PyCapsule
            A valid PyCapsule with name 'arrow_device_array' containing an
            ArrowDeviceArray pointer.
        """
        cdef:
            CSchema out_schema
            CDeviceArray out

        out_schema = CSchema._import_from_c_capsule(schema_capsule)
        out = CDeviceArray(
            device_array_capsule,
            <uintptr_t>PyCapsule_GetPointer(device_array_capsule, 'arrow_device_array'),
            out_schema
        )

        return out

    def __repr__(self) -> str:
        return _repr_utils.device_array_repr(self)



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

import _cython_3_0_11
import nanoarrow._device
from _typeshed import Incomplete
from nanoarrow._device import DeviceType as DeviceType
from typing import ClassVar

DEVICE_CPU: nanoarrow._device.Device
__reduce_cython__: _cython_3_0_11.cython_function_or_method
__setstate_cython__: _cython_3_0_11.cython_function_or_method
__test__: dict

class CArray:
    __pyx_vtable__: ClassVar[PyCapsule] = ...
    buffers: Incomplete
    children: Incomplete
    device_id: Incomplete
    device_type: Incomplete
    device_type_id: Incomplete
    dictionary: Incomplete
    length: Incomplete
    n_buffers: Incomplete
    n_children: Incomplete
    null_count: Incomplete
    offset: Incomplete
    schema: Incomplete
    @classmethod
    def __init__(cls, *args, **kwargs) -> None:
        """Create and return a new object.  See help(type) for accurate signature."""
    @staticmethod
    def allocate(*args, **kwargs):
        """Allocate a released ArrowArray"""
    def child(self, *args, **kwargs): ...
    def is_valid(self, *args, **kwargs):
        """Check for a non-null and non-released underlying ArrowArray"""
    def view(self, *args, **kwargs):
        """Allocate a :class:`CArrayView` to access the buffers of this array"""
    def __arrow_c_array__(self, *args, **kwargs):
        """
        Get a pair of PyCapsules containing a C ArrowArray representation of the object.

        Parameters
        ----------
        requested_schema : PyCapsule | None
            A PyCapsule containing a C ArrowSchema representation of a requested
            schema. Not supported.

        Returns
        -------
        Tuple[PyCapsule, PyCapsule]
            A pair of PyCapsules containing a C ArrowSchema and ArrowArray,
            respectively.
        """
    def __getitem__(self, index):
        """Return self[key]."""
    def __len__(self) -> int:
        """Return len(self)."""
    def __reduce__(self): ...

class CArrayBuilder:
    @classmethod
    def __init__(cls, *args, **kwargs) -> None:
        """Create and return a new object.  See help(type) for accurate signature."""
    @staticmethod
    def allocate(*args, **kwargs):
        """Create a CArrayBuilder

        Allocates memory for an ArrowArray and populates it with nanoarrow's
        ArrowArray private_data/release callback implementation. This should
        usually be followed by :meth:`init_from_type` or :meth:`init_from_schema`.
        """
    def append_bytes(self, *args, **kwargs): ...
    def append_strings(self, *args, **kwargs): ...
    def finish(self, *args, **kwargs):
        """Finish building this array

        Performs any steps required to return a valid ArrowArray and optionally
        validates the output to ensure that the result is valid (given the information
        the array has available to it).

        Parameters
        ----------
        validation_level : None, "full", "default", "minimal", or "none", optional
            Explicitly define a validation level or use None to perform default
            validation if possible. Validation may not be possible if children
            were set that were not created by nanoarrow.
        """
    def finish_device(self, *args, **kwargs):
        """Finish building this array and export to an ArrowDeviceArray

        Calls :meth:`finish`, propagating device information into an ArrowDeviceArray.
        """
    def init_from_schema(self, *args, **kwargs): ...
    def init_from_type(self, *args, **kwargs): ...
    def is_empty(self, *args, **kwargs):
        """Check if any items have been appended to this builder"""
    def resolve_null_count(self, *args, **kwargs):
        """Ensure the output null count is synchronized with existing buffers

        Note that this will not attempt to access non-CPU buffers such that
        :attr:`null_count` might still be -1 after calling this method.
        """
    def set_buffer(self, *args, **kwargs):
        """Set an ArrowArray buffer

        Sets a buffer of this ArrowArray such the pointer at array->buffers[i] is
        equal to buffer->data and such that the buffer's lifcycle is managed by
        the array. If move is True, the input Python object that previously wrapped
        the ArrowBuffer will be invalidated, which is usually the desired behaviour
        if you built or imported a buffer specifically to build this array. If move
        is False (the default), this function will a make a shallow copy via another
        layer of Python object wrapping.
        """
    def set_child(self, *args, **kwargs):
        """Set an ArrowArray child

        Set a child of this array by performing a show copy or optionally
        transferring ownership to this object. The initialized child array
        must have been initialized before this call by initializing this
        builder with a schema containing the correct number of children.
        """
    def set_length(self, *args, **kwargs): ...
    def set_null_count(self, *args, **kwargs): ...
    def set_offset(self, *args, **kwargs): ...
    def start_appending(self, *args, **kwargs):
        """Use append mode for building this ArrowArray

        Calling this method is required to produce a valid array prior to calling
        :meth:`append_strings` or `append_bytes`.
        """
    def __reduce__(self): ...

class CArrayView:
    buffers: Incomplete
    children: Incomplete
    dictionary: Incomplete
    layout: Incomplete
    length: Incomplete
    n_buffers: Incomplete
    n_children: Incomplete
    null_count: Incomplete
    offset: Incomplete
    storage_type: Incomplete
    storage_type_id: Incomplete
    @classmethod
    def __init__(cls, *args, **kwargs) -> None:
        """Create and return a new object.  See help(type) for accurate signature."""
    def buffer(self, *args, **kwargs): ...
    def buffer_type(self, *args, **kwargs): ...
    def child(self, *args, **kwargs): ...
    @staticmethod
    def from_array(*args, **kwargs): ...
    @staticmethod
    def from_schema(*args, **kwargs): ...
    def __len__(self) -> int:
        """Return len(self)."""
    def __reduce__(self): ...

class CDeviceArray:
    array: Incomplete
    device_id: Incomplete
    device_type: Incomplete
    device_type_id: Incomplete
    schema: Incomplete
    @classmethod
    def __init__(cls, *args, **kwargs) -> None:
        """Create and return a new object.  See help(type) for accurate signature."""
    def view(self, *args, **kwargs): ...
    def __arrow_c_array__(self, *args, **kwargs): ...
    def __arrow_c_device_array__(self, *args, **kwargs): ...
    def __reduce__(self): ...



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

# cython: language_level = 3

from libc.stdint cimport int32_t, int64_t, uintptr_t
from cpython.bytes cimport PyBytes_FromStringAndSize, PyBytes_AsString, PyBytes_Size
from cpython.pycapsule cimport PyCapsule_GetPointer

from nanoarrow_c cimport (
    ARROW_FLAG_DICTIONARY_ORDERED,
    ARROW_FLAG_MAP_KEYS_SORTED,
    ARROW_FLAG_NULLABLE,
    ArrowFree,
    ArrowLayout,
    ArrowMalloc,
    ArrowMetadataBuilderAppend,
    ArrowMetadataBuilderInit,
    ArrowMetadataReaderInit,
    ArrowMetadataReaderRead,
    ArrowSchema,
    ArrowSchemaAllocateChildren,
    ArrowSchemaAllocateDictionary,
    ArrowSchemaDeepCopy,
    ArrowSchemaInit,
    ArrowSchemaMove,
    ArrowSchemaRelease,
    ArrowSchemaSetMetadata,
    ArrowSchemaSetType,
    ArrowSchemaSetTypeDateTime,
    ArrowSchemaSetTypeDecimal,
    ArrowSchemaSetTypeFixedSize,
    ArrowSchemaSetFormat,
    ArrowSchemaSetName,
    ArrowSchemaToString,
    ArrowSchemaViewInit,
    ArrowStringView,
    ArrowTimeUnit,
    ArrowTimeUnitString,
    ArrowType,
    ArrowTypeString,
    NANOARROW_BUFFER_TYPE_NONE,
    NANOARROW_MAX_FIXED_BUFFERS,
    NANOARROW_TIME_UNIT_SECOND,
    NANOARROW_TIME_UNIT_MILLI,
    NANOARROW_TIME_UNIT_MICRO,
    NANOARROW_TIME_UNIT_NANO,
)

from nanoarrow cimport _types
from nanoarrow._buffer cimport CBuffer
from nanoarrow._utils cimport alloc_c_schema, Error

from typing import Iterable, List, Mapping, Tuple, Union

from nanoarrow import _repr_utils


# This is likely a better fit for a dedicated testing module; however, we need
# it here to produce nice error messages when ensuring that one or
# more arrays conform to a given or inferred schema.
cpdef assert_type_equal(actual, expected, bint check_nullability):
    """Test two schemas for data type equality

    Checks two CSchema objects for type equality (i.e., that an array with
    schema ``actual`` contains elements with the same logical meaning as and
    array with schema ``expected``). Notably, this excludes metadata from
    all nodes in the schema.

    Parameters
    ----------
    actual : CSchema
        The schema to be tested for equality
    expected : CSchema
        The schema against which to test
    check_nullability : bool
        If True, actual and expected will be considered equal if their
        data type information and marked nullability are identical.
    """
    if not isinstance(actual, CSchema):
        raise TypeError(f"actual is {type(actual).__name__}, not CSchema")

    if not isinstance(expected, CSchema):
        raise TypeError(f"expected is {type(expected).__name__}, not CSchema")

    if not actual.type_equals(expected, check_nullability=check_nullability):
        actual_label = actual._to_string(max_chars=80, recursive=True)
        expected_label = expected._to_string(max_chars=80, recursive=True)
        raise ValueError(
            f"Expected schema\n  '{expected_label}'"
            f"\nbut got\n  '{actual_label}'"
        )


cdef class CArrowTimeUnit:
    """
    Wrapper around ArrowTimeUnit to provide implementations in Python access
    to the values.
    """

    SECOND = NANOARROW_TIME_UNIT_SECOND
    MILLI = NANOARROW_TIME_UNIT_MILLI
    MICRO = NANOARROW_TIME_UNIT_MICRO
    NANO = NANOARROW_TIME_UNIT_NANO


cdef class CLayout:
    """Abstract buffer information for Arrow types

    Provides accessors for buffer counts, types, and attributes.
    """

    def __cinit__(self, base, uintptr_t ptr):
        self._base = base
        self._layout = <ArrowLayout*>ptr

        self._n_buffers = NANOARROW_MAX_FIXED_BUFFERS
        for i in range(NANOARROW_MAX_FIXED_BUFFERS):
            if self._layout.buffer_type[i] == NANOARROW_BUFFER_TYPE_NONE:
                self._n_buffers = i
                break

    @property
    def n_buffers(self) -> int:
        return self._n_buffers

    @property
    def buffer_data_type_id(self) -> int:
        return tuple(self._layout.buffer_data_type[i] for i in range(self._n_buffers))

    @property
    def element_size_bits(self) -> int:
        return tuple(self._layout.element_size_bits[i] for i in range(self._n_buffers))

    @property
    def child_size_elements(self) -> int:
        return self._layout.child_size_elements


cdef class SchemaMetadata:
    """Dictionary-like wrapper around a lazily-parsed CSchema.metadata string

    The Arrow C Data interface encodes key/value metadata as a bytes-to-bytes
    mapping using a specific packed binary encoding. This class maintains a
    reference to the underlying storage and parses it as required. Note that
    unlike a Python dictionary, ``SchemaMetadata`` can contain duplicate
    keys.
    """

    def __cinit__(self, object base, uintptr_t ptr):
        self._base = base
        self._metadata = <const char*>ptr

    @staticmethod
    def empty():
        """Create an empty SchemaMetadata with no keys or values"""
        return SchemaMetadata(None, 0)

    cdef _init_reader(self):
        cdef int code = ArrowMetadataReaderInit(&self._reader, self._metadata)
        Error.raise_error_not_ok("ArrowMetadataReaderInit()", code)

    def __len__(self):
        self._init_reader()
        return self._reader.remaining_keys

    def __contains__(self, item):
        for key, _ in self.items():
            if item == key:
                return True

        return False

    def __getitem__(self, k) -> bytes:
        """Get the value associated with a unique key

        Retrieves the unique value associated with k. Raises KeyError if
        k does not point to exactly one value in the metadata.
        """
        out = None

        for key, value in self.items():
            if k == key:
                if out is None:
                    out = value
                else:
                    raise KeyError(f"key {k} matches more than one value in metadata")

        if out is None:
            raise KeyError(f"Key {k} not found")

        return out

    def __iter__(self):
        for key, _ in self.items():
            yield key

    def keys(self) -> List[bytes]:
        """List meadata keys

        The result may contain duplicate keys if they exist in the metadata.
        """
        return list(self)

    def values(self) -> List[bytes]:
        """List metadata values"""
        return [value for _, value in self.items()]

    def items(self) -> Iterable[bytes, bytes]:
        """Iterate over key/value pairs

        The result may contain duplicate keys if they exist in the metadata."""
        cdef ArrowStringView key
        cdef ArrowStringView value
        self._init_reader()
        while self._reader.remaining_keys > 0:
            ArrowMetadataReaderRead(&self._reader, &key, &value)
            key_obj = PyBytes_FromStringAndSize(key.data, key.size_bytes)
            value_obj = PyBytes_FromStringAndSize(value.data, value.size_bytes)
            yield key_obj, value_obj

    def __repr__(self) -> str:
        lines = [
            f"<{_repr_utils.make_class_label(self)}>",
            _repr_utils.metadata_repr(self)
        ]
        return "\n".join(lines)


cdef class CSchema:
    """Low-level ArrowSchema wrapper

    This object is a literal wrapper around a read-only ArrowSchema. It provides field accessors
    that return Python objects and handles the C Data interface lifecycle (i.e., initialized
    ArrowSchema structures are always released).

    See ``nanoarrow.c_schema()`` for construction and usage examples.
    """

    @staticmethod
    def allocate() -> CSchema:
        """Allocate a released CSchema"""
        cdef ArrowSchema* c_schema_out
        base = alloc_c_schema(&c_schema_out)
        return CSchema(base, <uintptr_t>(c_schema_out))

    def __cinit__(self, object base, uintptr_t addr):
        self._base = base
        self._ptr = <ArrowSchema*>addr

    def __deepcopy__(self, memo=None) -> CSchema:
        cdef CSchema out = CSchema.allocate()
        cdef int code = ArrowSchemaDeepCopy(self._ptr, out._ptr)
        Error.raise_error_not_ok("ArrowSchemaDeepCopy()", code)

        return out

    @staticmethod
    def _import_from_c_capsule(schema_capsule) -> CSchema:
        """Import from a ArrowSchema PyCapsule

        Parameters
        ----------
        schema_capsule : PyCapsule
            A valid PyCapsule with name 'arrow_schema' containing an
            ArrowSchema pointer.
        """
        return CSchema(
            schema_capsule,
            <uintptr_t>PyCapsule_GetPointer(schema_capsule, "arrow_schema")
        )

    def __arrow_c_schema__(self):
        """
        Export to a ArrowSchema PyCapsule
        """
        self._assert_valid()

        cdef ArrowSchema* c_schema_out
        schema_capsule = alloc_c_schema(&c_schema_out)

        cdef int code = ArrowSchemaDeepCopy(self._ptr, c_schema_out)
        Error.raise_error_not_ok("ArrowSchemaDeepCopy", code)
        return schema_capsule

    @property
    def _capsule(self):
        """
        Returns the capsule backing this CSchema or None if it does not exist
        or points to a parent ArrowSchema.
        """
        cdef ArrowSchema* maybe_capsule_ptr
        maybe_capsule_ptr = <ArrowSchema*>PyCapsule_GetPointer(self._base, 'arrow_schema')

        # This will return False if this is a child CSchema whose capsule holds
        # the parent ArrowSchema
        if maybe_capsule_ptr == self._ptr:
            return self._base

        return None

    def _addr(self) -> int:
        return <uintptr_t>self._ptr

    def is_valid(self) -> bool:
        """Check for a non-null and non-released underlying ArrowSchema"""
        return self._ptr != NULL and self._ptr.release != NULL

    def _assert_valid(self):
        if self._ptr == NULL:
            raise RuntimeError("schema is NULL")
        if self._ptr.release == NULL:
            raise RuntimeError("schema is released")

    def _to_string(self, int64_t max_chars=0, recursive=False) -> str:
        cdef int64_t n_chars
        if max_chars == 0:
            n_chars = ArrowSchemaToString(self._ptr, NULL, 0, recursive)
        else:
            n_chars = max_chars

        cdef char* out = <char*>ArrowMalloc(n_chars + 1)
        if not out:
            raise MemoryError()

        ArrowSchemaToString(self._ptr, out, n_chars + 1, recursive)
        out_str = out.decode("UTF-8")
        ArrowFree(out)

        return out_str

    def __repr__(self) -> str:
        return _repr_utils.schema_repr(self)

    def type_equals(self, CSchema other, check_nullability: bool=False) -> bool:
        """Test two schemas for data type equality

        Checks two CSchema objects for type equality (i.e., that an array with
        schema ``actual`` contains elements with the same logical meaning as and
        array with schema ``expected``). Notably, this excludes metadata from
        all nodes in the schema.

        Parameters
        ----------
        other : CSchema
            The schema against which to test
        check_nullability : bool
            If True, actual and expected will be considered equal if their
            data type information and marked nullability are identical.
        """
        self._assert_valid()

        if self._ptr == other._ptr:
            return True

        if self.format != other.format:
            return False

        # Nullability is not strictly part of the "type"; however, performing
        # this check recursively is verbose to otherwise accomplish and
        # sometimes this does matter.
        cdef int64_t flags = self.flags
        cdef int64_t other_flags = other.flags
        if not check_nullability:
            flags &= ~ARROW_FLAG_NULLABLE
            other_flags &= ~ARROW_FLAG_NULLABLE

        if flags != other_flags:
            return False

        if self.n_children != other.n_children:
            return False

        for child, other_child in zip(self.children, other.children):
            if not child.type_equals(other_child, check_nullability=check_nullability):
                return False

        if (self.dictionary is None) != (other.dictionary is None):
            return False

        if self.dictionary is not None:
            if not self.dictionary.type_equals(
                other.dictionary,
                check_nullability=check_nullability
            ):
                return False

        return True


    @property
    def format(self) -> str:
        self._assert_valid()
        if self._ptr.format != NULL:
            return self._ptr.format.decode()

    @property
    def name(self) -> Union[str, None]:
        self._assert_valid()
        if self._ptr.name != NULL:
            return self._ptr.name.decode()
        else:
            return None

    @property
    def flags(self) -> int:
        return self._ptr.flags

    @property
    def metadata(self) -> SchemaMetadata:
        self._assert_valid()
        if self._ptr.metadata != NULL:
            return SchemaMetadata(self._base, <uintptr_t>self._ptr.metadata)
        else:
            return None

    @property
    def n_children(self) -> int:
        self._assert_valid()
        return self._ptr.n_children

    def child(self, int64_t i):
        self._assert_valid()
        if i < 0 or i >= self._ptr.n_children:
            raise IndexError(f"{i} out of range [0, {self._ptr.n_children})")

        return CSchema(self._base, <uintptr_t>self._ptr.children[i])

    @property
    def children(self) -> Iterable[CSchema]:
        for i in range(self.n_children):
            yield self.child(i)

    @property
    def dictionary(self) -> Union[CSchema, None]:
        self._assert_valid()
        if self._ptr.dictionary != NULL:
            return CSchema(self, <uintptr_t>self._ptr.dictionary)
        else:
            return None

    def modify(self, *, format=None, name=None, flags=None, nullable=None,
               metadata=None, children=None, dictionary=None, validate=True) -> CSchema:
        cdef CSchemaBuilder builder = CSchemaBuilder.allocate()

        if format is None:
            builder.set_format(self.format)
        else:
            builder.set_format(format)

        if name is None:
            builder.set_name(self.name)
        elif name is not False:
            builder.set_name(name)

        if flags is None:
            builder.set_flags(self.flags)
        else:
            builder.set_flags(flags)

        if nullable is not None:
            builder.set_nullable(nullable)

        if metadata is None:
            if self.metadata is not None:
                builder.append_metadata(self.metadata)
        else:
            builder.append_metadata(metadata)

        if children is None:
            if self.n_children > 0:
                builder.allocate_children(self.n_children)
                for i, child in enumerate(self.children):
                    builder.set_child(i, None, child)
        elif hasattr(children, "items"):
            builder.allocate_children(len(children))
            for i, item in enumerate(children.items()):
                name, child = item
                builder.set_child(i, name, child)
        else:
            builder.allocate_children(len(children))
            for i, child in enumerate(children):
                builder.set_child(i, None, child)

        if dictionary is None:
            if self.dictionary:
                builder.set_dictionary(self.dictionary)
        elif dictionary is not False:
            builder.set_dictionary(dictionary)

        if validate:
            builder.validate()

        return builder.finish()


cdef class CSchemaView:
    """Low-level ArrowSchemaView wrapper

    This object is a literal wrapper around a read-only ArrowSchemaView. It provides field accessors
    that return Python objects and handles structure lifecycle. Compared to an ArrowSchema,
    the nanoarrow ArrowSchemaView facilitates access to the deserialized content of an ArrowSchema
    (e.g., parameter values for parameterized types).

    See `nanoarrow.c_schema_view()` for construction and usage examples.
    """

    def __cinit__(self, CSchema schema):
        self._base = schema
        self._schema_view.type = <ArrowType>_types.UNINITIALIZED
        self._schema_view.storage_type = <ArrowType>_types.UNINITIALIZED

        cdef Error error = Error()
        cdef int code = ArrowSchemaViewInit(&self._schema_view, schema._ptr, &error.c_error)
        error.raise_message_not_ok("ArrowSchemaViewInit()", code)

        self._dictionary_ordered = schema._ptr.flags & ARROW_FLAG_DICTIONARY_ORDERED
        self._nullable = schema._ptr.flags & ARROW_FLAG_NULLABLE
        self._map_keys_sorted = schema._ptr.flags & ARROW_FLAG_MAP_KEYS_SORTED

    @property
    def layout(self) -> CLayout:
        return CLayout(self, <uintptr_t>&self._schema_view.layout)

    @property
    def type_id(self) -> int:
        return self._schema_view.type

    @property
    def storage_type_id(self) -> int:
        return self._schema_view.storage_type

    @property
    def storage_buffer_format(self) -> Union[str, None]:
        if self.buffer_format is not None:
            return self.buffer_format
        elif _types.equal(self._schema_view.type, _types.DATE32):
            return 'i'
        elif _types.one_of(
            self._schema_view.type,
            (_types.TIMESTAMP, _types.DATE64, _types.DURATION)
        ):
            return 'q'
        elif self.extension_name:
            return self._get_buffer_format()
        else:
            return None

    @property
    def buffer_format(self) -> Union[str, None]:
        """The Python struct format representing an element of this type
        or None if there is no Python format string that can represent this
        type without loosing information.
        """
        if self.extension_name:
            return None
        else:
            return self._get_buffer_format()

    def _get_buffer_format(self):
        if self._schema_view.type != self._schema_view.storage_type:
            return None

        # String/binary types do not have format strings as far as the Python
        # buffer protocol is concerned
        if self.layout.n_buffers != 2:
            return None

        cdef char out[128]
        cdef int element_size_bits = 0
        if _types.equal(self._schema_view.type, _types.FIXED_SIZE_BINARY):
            element_size_bits = self._schema_view.fixed_size * 8

        try:
            _types.to_format(self._schema_view.type, element_size_bits, sizeof(out), out)
            return out.decode()
        except ValueError:
            return None

    @property
    def type(self) -> str:
        cdef const char* type_str = ArrowTypeString(self._schema_view.type)
        if type_str != NULL:
            return type_str.decode()
        else:
            raise ValueError("ArrowTypeString() returned NULL")

    @property
    def storage_type(self) -> str:
        cdef const char* type_str = ArrowTypeString(self._schema_view.storage_type)
        if type_str != NULL:
            return type_str.decode()
        else:
            raise ValueError("ArrowTypeString() returned NULL")

    @property
    def dictionary_ordered(self) -> Union[bool, None]:
        if _types.equal(self._schema_view.type, _types.DICTIONARY):
            return self._dictionary_ordered != 0
        else:
            return None

    @property
    def nullable(self) -> bool:
        return self._nullable != 0

    @property
    def map_keys_sorted(self) -> Union[bool, None]:
        if _types.equal(self._schema_view.type, _types.MAP):
            return self._map_keys_sorted != 0
        else:
            return None

    @property
    def fixed_size(self) -> Union[bool, None]:
        if _types.is_fixed_size(self._schema_view.type):
            return self._schema_view.fixed_size
        else:
            return None

    @property
    def decimal_bitwidth(self) -> Union[int, None]:
        if _types.is_decimal(self._schema_view.type):
            return self._schema_view.decimal_bitwidth
        else:
            return None

    @property
    def decimal_precision(self) -> Union[int, None]:
        if _types.is_decimal(self._schema_view.type):
            return self._schema_view.decimal_precision
        else:
            return None

    @property
    def decimal_scale(self) -> Union[int, None]:
        if _types.is_decimal(self._schema_view.type):
            return self._schema_view.decimal_scale
        else:
            return None

    @property
    def time_unit_id(self) -> Union[int, None]:
        if _types.has_time_unit(self._schema_view.type):
            return self._schema_view.time_unit
        else:
            return None

    @property
    def time_unit(self) -> Union[str, None]:
        if _types.has_time_unit(self._schema_view.type):
            return ArrowTimeUnitString(self._schema_view.time_unit).decode()
        else:
            return None

    @property
    def timezone(self) -> Union[str, None]:
        if _types.equal(self._schema_view.type, _types.TIMESTAMP):
            return self._schema_view.timezone.decode()
        else:
            return None

    @property
    def union_type_ids(self) -> Union[Tuple[int, ...], None]:
        if _types.is_union(self._schema_view.type):
            type_ids_str = self._schema_view.union_type_ids.decode().split(',')
            return (int(type_id) for type_id in type_ids_str)
        else:
            return None

    @property
    def extension_name(self) -> Union[str, None]:
        if self._schema_view.extension_name.data != NULL:
            name_bytes = PyBytes_FromStringAndSize(
                self._schema_view.extension_name.data,
                self._schema_view.extension_name.size_bytes
            )
            return name_bytes.decode()
        else:
            return None

    @property
    def extension_metadata(self) -> Union[bytes, None]:
        if self._schema_view.extension_name.data != NULL:
            return PyBytes_FromStringAndSize(
                self._schema_view.extension_metadata.data,
                self._schema_view.extension_metadata.size_bytes
            )
        else:
            return None

    def __repr__(self) -> str:
        return _repr_utils.schema_view_repr(self)


cdef class CSchemaBuilder:
    """Helper for constructing an ArrowSchema

    The primary function of this class is to wrap the nanoarrow C library calls
    that build up the components of an ArrowSchema.
    """

    def __cinit__(self, CSchema schema):
        self.c_schema = schema
        self._ptr = schema._ptr
        if self._ptr.release == NULL:
            ArrowSchemaInit(self._ptr)

    @staticmethod
    def allocate() -> CSchemaBuilder:
        """Create a CSchemaBuilder

        Allocates memory for an ArrowSchema and populates it with nanoarrow's
        ArrowSchema private_data/release callback implementation. This should
        usually be followed by :meth:`set_type` or :meth:`set_format`.
        """
        return CSchemaBuilder(CSchema.allocate())

    def append_metadata(self, metadata: Mapping[bytes, bytes]) -> CSchemaBuilder:
        """Append key/value metadata"""
        cdef CBuffer buffer = CBuffer.empty()

        cdef const char* existing_metadata = self.c_schema._ptr.metadata
        cdef int code = ArrowMetadataBuilderInit(buffer._ptr, existing_metadata)
        Error.raise_error_not_ok("ArrowMetadataBuilderInit()", code)

        cdef ArrowStringView key
        cdef ArrowStringView value
        cdef int32_t keys_added = 0

        for k, v in metadata.items():
            k = k.encode() if isinstance(k, str) else bytes(k)
            key.data = PyBytes_AsString(k)
            key.size_bytes = PyBytes_Size(k)

            v = v.encode() if isinstance(v, str) else bytes(v)
            value.data = PyBytes_AsString(v)
            value.size_bytes = PyBytes_Size(v)

            code = ArrowMetadataBuilderAppend(buffer._ptr, key, value)
            Error.raise_error_not_ok("ArrowMetadataBuilderAppend()", code)

            keys_added += 1

        if keys_added > 0:
            code = ArrowSchemaSetMetadata(self.c_schema._ptr, <const char*>buffer._ptr.data)
            Error.raise_error_not_ok("ArrowSchemaSetMetadata()", code)

        return self

    def child(self, int64_t i) -> CSchemaBuilder:
        return CSchemaBuilder(self.c_schema.child(i))

    def set_type(self, int type_id) -> CSchemaBuilder:
        self.c_schema._assert_valid()

        cdef int code = ArrowSchemaSetType(self._ptr, <ArrowType>type_id)
        Error.raise_error_not_ok("ArrowSchemaSetType()", code)

        return self

    def set_type_decimal(self, int type_id, int precision, int scale) -> CSchemaBuilder:
        self.c_schema._assert_valid()

        cdef int code = ArrowSchemaSetTypeDecimal(self._ptr, <ArrowType>type_id, precision, scale)
        Error.raise_error_not_ok("ArrowSchemaSetType()", code)

    def set_type_fixed_size(self, int type_id, int fixed_size) -> CSchemaBuilder:
        self.c_schema._assert_valid()

        cdef int code = ArrowSchemaSetTypeFixedSize(self._ptr, <ArrowType>type_id, fixed_size)
        Error.raise_error_not_ok("ArrowSchemaSetTypeFixedSize()", code)

        return self

    def set_type_date_time(self, int type_id, int time_unit, timezone) -> CSchemaBuilder:
        self.c_schema._assert_valid()

        cdef int code
        if timezone is None:
            code = ArrowSchemaSetTypeDateTime(self._ptr, <ArrowType>type_id, <ArrowTimeUnit>time_unit, NULL)
        else:
            timezone = str(timezone)
            code = ArrowSchemaSetTypeDateTime(self._ptr, <ArrowType>type_id, <ArrowTimeUnit>time_unit, timezone.encode("UTF-8"))

        Error.raise_error_not_ok("ArrowSchemaSetTypeDateTime()", code)

        return self

    def set_format(self, str format) -> CSchemaBuilder:
        self.c_schema._assert_valid()

        cdef int code = ArrowSchemaSetFormat(self._ptr, format.encode("UTF-8"))
        Error.raise_error_not_ok("ArrowSchemaSetFormat()", code)

        return self

    def set_name(self, name) -> CSchemaBuilder:
        self.c_schema._assert_valid()

        cdef int code
        if name is None:
            code = ArrowSchemaSetName(self._ptr, NULL)
        else:
            name = str(name)
            code = ArrowSchemaSetName(self._ptr, name.encode("UTF-8"))

        Error.raise_error_not_ok("ArrowSchemaSetName()", code)

        return self

    def allocate_children(self, int n) -> CSchemaBuilder:
        self.c_schema._assert_valid()

        cdef int code = ArrowSchemaAllocateChildren(self._ptr, n)
        Error.raise_error_not_ok("ArrowSchemaAllocateChildren()", code)

        return self

    def set_child(self, int64_t i, name, CSchema child_src) -> CSchemaBuilder:
        self.c_schema._assert_valid()

        if i < 0 or i >= self._ptr.n_children:
            raise IndexError(f"Index out of range: {i}")

        if self._ptr.children[i].release != NULL:
            ArrowSchemaRelease(self._ptr.children[i])

        cdef int code = ArrowSchemaDeepCopy(child_src._ptr, self._ptr.children[i])
        Error.raise_error_not_ok("ArrowSchemaDeepCopy()", code)

        if name is not None:
            name = str(name)
            code = ArrowSchemaSetName(self._ptr.children[i], name.encode("UTF-8"))
            Error.raise_error_not_ok("ArrowSchemaSetName()", code)

        return self

    def set_dictionary(self, CSchema dictionary) -> CSchemaBuilder:
        self.c_schema._assert_valid()

        cdef int code
        if self._ptr.dictionary == NULL:
            code = ArrowSchemaAllocateDictionary(self._ptr)
            Error.raise_error_not_ok("ArrowSchemaAllocateDictionary()", code)

        if self._ptr.dictionary.release != NULL:
            ArrowSchemaRelease(self._ptr.dictionary)

        code = ArrowSchemaDeepCopy(dictionary._ptr, self._ptr.dictionary)
        Error.raise_error_not_ok("ArrowSchemaDeepCopy()", code)

        return self

    def set_flags(self, flags) -> CSchemaBuilder:
        self._ptr.flags = flags
        return self

    def set_nullable(self, nullable) -> CSchemaBuilder:
        if nullable:
            self._ptr.flags = self._ptr.flags | ARROW_FLAG_NULLABLE
        else:
            self._ptr.flags = self._ptr.flags & ~ARROW_FLAG_NULLABLE

        return self

    def set_dictionary_ordered(self, dictionary_ordered) -> CSchemaBuilder:
        if dictionary_ordered:
            self._ptr.flags = self._ptr.flags | ARROW_FLAG_DICTIONARY_ORDERED
        else:
            self._ptr.flags = self._ptr.flags & ~ARROW_FLAG_DICTIONARY_ORDERED

        return self

    def set_map_keys_sorted(self, map_keys_sorted) -> CSchemaBuilder:
        if map_keys_sorted:
            self._ptr.flags = self._ptr.flags | ARROW_FLAG_MAP_KEYS_SORTED
        else:
            self._ptr.flags = self._ptr.flags & ~ARROW_FLAG_MAP_KEYS_SORTED

        return self

    def validate(self) -> CSchemaView:
        return CSchemaView(self.c_schema)

    def finish(self) -> CSchema:
        self.c_schema._assert_valid()
        cdef CSchema out = CSchema.allocate()
        ArrowSchemaMove(self.c_schema._ptr, out._ptr)
        ArrowSchemaInit(self.c_schema._ptr)
        return out



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

# cython: language_level = 3

from libc.stdint cimport int64_t

from nanoarrow_c cimport (
    ArrowBuffer,
    ArrowBufferView,
    ArrowType,
)

from nanoarrow._device cimport Device, CSharedSyncEvent


cdef class CBufferView:
    cdef object _base
    cdef ArrowBufferView _ptr
    cdef ArrowType _data_type
    cdef CSharedSyncEvent _event
    cdef Py_ssize_t _element_size_bits
    cdef Py_ssize_t _shape
    cdef Py_ssize_t _strides
    cdef int64_t _n_elements
    cdef char _format[128]

    cdef _check_copy_into_bounds(self, Py_buffer* dest, int64_t offset, int64_t length,
                                 int64_t dest_offset, int64_t dest_itemsize)

    cdef Py_ssize_t _item_size(self)

    cdef _do_getbuffer(self, Py_buffer *buffer, int flags)

    cdef _do_releasebuffer(self, Py_buffer* buffer)

cdef class CBuffer:
    cdef object _base
    cdef ArrowBuffer* _ptr
    cdef ArrowType _data_type
    cdef int _element_size_bits
    cdef char _format[32]
    cdef Device _device
    cdef CBufferView _view
    cdef int _get_buffer_count

    cdef _assert_valid(self)

    cdef _assert_buffer_count_zero(self)

    cdef _populate_view(self)
